/* $Header: /s/misc/Program Files/Hummingbird/Connectivity/9.00/NFSClient/utility/hcllock.c 3     1/06/04 1:28p Frisch $ */
#define VERSION  "10.0.0.2"
/*
HCLNFSD (formerly BWNFSD) is an RPC daemon used in conjunction with
HCLNFS (a client NFS for DOS based PCs). HCLNFSD provides Authentication,
Print Spooling, DOS 3.1 Locking, DOS 3.1 Sharing, GID name mapping,
UID name mapping services for HCLNFS and associated applications on the PC.
HCLNFSD is being used also by Macintosh NFS clients.

The HCLNFSD code is originally copyright Beame & Whiteside Software Ltd.
and now is released to the Public Domain. The intent is that all server vendors
included a version of HCLNFSD with their operating system. HCLNFSD can run
simultanteously with PCNFSD, but provides more features.

	Please send modifications to:

		Hummingbird Ltd.
		1 Sparks Avenue
		North York, Ontario
		Canada	M2H 2W1
		+1 (416) 496-2200
		support@hummingbird.com

Please modify by including "ifdefs" for your particular operating
system, with appropriate modifications to the "makefile".

HCLLOCK.C provides the locking and sharing service for HCLNFS. This should be
used if rpc.lockd does not include the extensions for DOS 3.1 locking or
sharing, or if your system has the regular distribution of the rpc.lockd as
the extensions most likely will not work.

One might want to make operating system calls to lock and unlock the files
so that locking occurs not only against other HCLNFS or BWNFS clients, but
native clients as well.

MODIFICATION HISTORY
--------------------
31/07/92 fjw Cleaned up the code a little
01/08/92 fjw Merged in SVR4 code from Frank Glass (gatech!holos0!fsg)
19/10/92 fjw Changed code to use some manifest constants
05/11/92 fjw Punched up some of the debug output
17/12/93 fjw Changed some IFDEF names for clarity
09/02/94 fjw Added code to allow locks/file/process to be removed
07/24/95 fjw Rewrite to enforce 32-bit types for OSF/1; molto source changes
09/13/95 fjw Change code to allow program name to be user-selectable
10/24/95 cfb Fix 9/2/94 mod to require locks removed be owned by PC unlocking
02/12/97 fjw Made the cookie cache larger than 1, to handle multiple requests
04/01/97 cfb added exclusive and read shared locking
06/12/97 fjw added code to help with dangling locks
*/

#define PROGNAME "hclnfsd"

/*
 * HASH_LOCATION is a number between 0 and FHSIZE-1 inclusive. It should be
 * set to the position in the file handle FHSIZE byte structure which
 * changes the most between different files.
 */
#define HASH_LOCATION	17

#define LOCK_SHARE	20
#define LOCK_UNSHARE	21
#define LOCK_LOCK	22
#define LOCK_REMOVE_ALL 23
#define LOCK_UNLOCK	24
#define LOCK_BREAK_FH	101		/* break all locks held on FH */
#define LOCK_BREAK_IP	102		/* break all locks held by IP */
#define LOCK_ZAP_LOCKS	105		/* remove locks on file by process */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#ifdef USE_BCOPY
#define memcmp(a,b,c) bcmp((b),(a),(c))
#define memcpy(a,b,c) bcopy((b),(a),(c))
#else
#include <memory.h>
#endif

#ifndef INT32_LONG
typedef unsigned int _uint32;
#else
typedef unsigned long _uint32;
#endif

#ifdef INT_SIG
#define Void int
#define Return return(0)
#else
#define Void void
#define Return return
#endif

#define LOCK	'L'
#define SHARE	'S'

      /*	      1st Open		   */
      /* ALL	 DR	DW     DRW   COMP  */
      /* I O B	I O B  I O B  I O B  I O B */
char shares[15][15] = {
	{1,1,1, 0,0,0, 1,1,1, 0,0,0, 0,0,0}, /* I */
	{1,1,1, 1,1,1, 0,0,0, 0,0,0, 0,0,0}, /* O  ALL */
	{1,1,1, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* B */
	{0,1,0, 0,0,0, 0,1,0, 0,0,0, 0,0,0}, /* I */
	{0,1,0, 0,1,0, 0,0,0, 0,0,0, 0,0,0}, /* O  DR */
	{0,1,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* B */
	{1,0,0, 0,0,0, 1,0,0, 0,0,0, 0,0,0}, /* I */
	{1,0,0, 1,0,0, 0,0,0, 0,0,0, 0,0,0}, /* O DW 2nd Open */
	{1,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* B */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* I */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* O  DRW */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0}, /* B */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 1,1,1}, /* I */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 1,1,1}, /* O  COMP */
	{0,0,0, 0,0,0, 0,0,0, 0,0,0, 1,1,1}  /* B */
};

struct LK {
	char type;
	struct LK *nxt;
	struct LK *prv;
	struct LK *nxt_fh;
	struct LK *prv_fh;
	struct FH *fh;
	struct PC *pc;
	unsigned long num_failed;
	union {
		struct {
			char mode;
			char access;
		} s;
		struct {
			_uint32 offset;
			_uint32 length;
			_uint32 lockowner;
			_uint32 exclusive;
		} l;
	} lock;
};

#define FHSIZE 32
struct FH {
	char fh[FHSIZE];
	struct LK *lock;
	struct FH *nxt;
	struct FH *prv;
};

#define LOCK_CACHE_SIZE 8
struct PC {
	char *owner;
	struct PC *nxt;
	struct PC *prv;
	struct LK *lock;
	struct cache {
		_uint32 last_cookie;
		_uint32 last_result;
	} cache[LOCK_CACHE_SIZE];
	int cache_pointer;
};

#define NUM_PCS 256
#define NUM_FHS NUM_PCS

struct PC *pcs[NUM_PCS];
struct FH *fhs[NUM_FHS];
extern int debugmode;

char *
requestname(req)
	int req;
{
static char buff[100];

	switch(req) {
	case LOCK_SHARE:
		strcpy(buff, "Share file");
		break;
	case LOCK_UNSHARE:
		strcpy(buff, "UnShare file");
		break;
	case LOCK_LOCK:
		strcpy(buff, "Lock file");
		break;
	case LOCK_REMOVE_ALL:
		strcpy(buff, "Remove all locks/shares held by owner");
		break;
	case LOCK_UNLOCK:
		strcpy(buff, "UnLock file");
		break;
	case LOCK_BREAK_FH:
		strcpy(buff, "Break all Locks/Shares on file");
		break;
	case LOCK_BREAK_IP:
		strcpy(buff, "Break all Locks/Shares held by owner's IP");
		break;
	case LOCK_ZAP_LOCKS:
		strcpy(buff, "Remove all locks on file by process");
		break;
	default:
		sprintf(buff, "Unknown request - %d", req);
		break;
	}
	return(buff);
}

char *
accessname(acc)
	_uint32 acc;
{
static char buff[30];

	switch(acc) {
	case 1:
		strcpy(buff, "Input");
		break;
	case 2:
		strcpy(buff, "Output");
		break;
	case 3:
		strcpy(buff, "Input/Output");
		break;
	default:
		sprintf(buff, "Unknown access value - %d", acc);
		break;
	}
	return(buff);
}

char *
modename(mode)
	_uint32 mode;
{
static char buff[30];

	switch(mode) {
	case 0:
		strcpy(buff, "Deny None");
		break;
	case 1:
		strcpy(buff, "Deny Read");
		break;
	case 2:
		strcpy(buff, "Deny Write");
		break;
	case 3:
		strcpy(buff, "Deny Read/Write");
		break;
	case 4:
		strcpy(buff, "Compatibility");
		break;
	default:
		sprintf(buff, "Unknown mode value - %d", mode);
		break;
	}
	return(buff);
}

char
*my_strdup(p)
	char *p;
{
char *p1;

	if ((p1=(char *)malloc(strlen(p)+1)) == NULL)
		return(NULL);
	return(strcpy(p1, p));
}

int
OwnerHash(owner)
	char *owner;
{
int len, ret;

	len = owner? strlen(owner): 0;
	if (len == 0)
		ret = 0;
	else if (len == 1)
		ret = *(owner+len-1) & 0xff;
	else
		ret = (*(owner+len-1) ^ *(owner+len-2)) & 0xff;
	return(ret);
}

Void
display_locks(unused)
	int unused;
{
int i, j, k;
struct PC *pc1;
struct LK *lk1;
struct FH *fh1;
FILE *lock_file;

#ifdef SIGUSR1
	(void) signal(SIGUSR1, display_locks);
#endif
	if ((lock_file=fopen("/etc/nfslocks.dmp", "w")) == NULL)
		Return;
	for (i=0; i<NUM_PCS; i++) {
		j = 0;
		pc1 = pcs[i];
		while (pc1 != NULL) {
			j += fprintf(lock_file, "%3d %s -> ", i, pc1->owner);
			lk1 = pc1->lock;
			while (lk1 != NULL) {
				j += fprintf(lock_file, "%c/%3u/", lk1->type, (lk1->fh)->fh[HASH_LOCATION] & 0xff);
				if (lk1->type == SHARE) {
					j += fprintf(lock_file, "%1d -> ", lk1->lock.s.mode);
				}
				else {
				unsigned long t1=lk1->lock.l.offset, t2=lk1->lock.l.length, t3=lk1->lock.l.lockowner;
					j += fprintf(lock_file, "%c%3ld+%3ld(%3ld) -> ", (lk1->lock.l.exclusive)? 'E': 'R', t1, t2, t3);
				}
				if (j > 70)
					j = fprintf(lock_file, "\n                ") - 1;
				lk1 = lk1->nxt;
			}
			fprintf(lock_file, "NULL\n");
			pc1 = pc1->nxt;
		}
	}
	fprintf(lock_file, "File Handles\n");
	for (i=0; i<NUM_FHS; i++) {
		j = 0;
		fh1 = fhs[i];
		while (fh1 != NULL) {
			j += fprintf(lock_file, "%3d ", i);
			for (k=0; k<FHSIZE; k++)
			       j += fprintf(lock_file, "%02X", fh1->fh[k] & 0xff);
			j += fprintf(lock_file, " -> ");
			lk1 = fh1->lock;
			while (lk1 != NULL) {
				j += fprintf(lock_file, "%c/%3u/", lk1->type, OwnerHash((lk1->pc)->owner));
				if (lk1->type == SHARE) {
					j += fprintf(lock_file, "%1d -> ", lk1->lock.s.mode);
				}
				else {
				unsigned long t1=lk1->lock.l.offset, t2=lk1->lock.l.length, t3=lk1->lock.l.lockowner;
					j += fprintf(lock_file, "%c%3ld+%3ld(%3ld) -> ", (lk1->lock.l.exclusive)? 'E': 'R', t1, t2, t3);
				}
				if (j > 70)
					j = fprintf(lock_file, "\n                ") - 1;
				lk1 = lk1->nxt_fh;
			}
			fprintf(lock_file, "NULL\n");
			fh1 = fh1->nxt;
		}
	}
	fclose(lock_file);
	Return;
}

Void
display_locks2(unused)
	int unused;
{
int i, j, k;
struct PC *pc1;
struct LK *lk1;
struct FH *fh1;
FILE *lock_file;

#ifdef SIGUSR2
	(void) signal(SIGUSR2, display_locks2);
#endif
	if ((lock_file=fopen("/etc/nfslocks2.dmp", "w")) == NULL)
		Return;
	fprintf(lock_file, "File Handles\n");
	for (i=0; i<NUM_FHS; i++) {
		j = 0;
		fh1 = fhs[i];
		while (fh1 != NULL) {
			j += fprintf(lock_file, "%3d ", i);
			for (k=0; k<FHSIZE; k++)
			       j += fprintf(lock_file, "%02X", fh1->fh[k] & 0xff);
			j += fprintf(lock_file, " -> ");
			lk1 = fh1->lock;
			while (lk1 != NULL) {
				j += fprintf(lock_file, "%c/%s/%lu/", lk1->type, (lk1->pc)->owner, lk1->num_failed);
				if (lk1->type == SHARE) {
					j += fprintf(lock_file, "%1d -> ", lk1->lock.s.mode);
				}
				else {
				unsigned long t1=lk1->lock.l.offset, t2=lk1->lock.l.length, t3=lk1->lock.l.lockowner;
					j += fprintf(lock_file, "%c%3ld+%3ld(%3ld) -> ", (lk1->lock.l.exclusive)? 'E': 'R', t1, t2, t3);
				}
				if (j > 70)
					j = fprintf(lock_file, "\n                ") - 1;
				lk1 = lk1->nxt_fh;
			}
			fprintf(lock_file, "NULL\n");
			fh1 = fh1->nxt;
		}
	}
	fclose(lock_file);
	Return;
}

void
init_locks()
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [init_locks] called\n", PROGNAME);
	for (i=0; i<NUM_PCS; i++)
		pcs[i] = NULL;
	for (i=0; i<NUM_FHS; i++)
		fhs[i] = NULL;
#ifdef SIGUSR1
	(void) signal(SIGUSR1, display_locks);
#endif
#ifdef SIGUSR2
	(void) signal(SIGUSR2, display_locks2);
#endif
}

#define TRUE  1
#define FALSE 0

static int pc_ptr, fh_ptr;

struct PC *
locate_by_owner(owner)
	char *owner;
{
struct PC *p;

	if (debugmode)
		fprintf(stdout, "%s: [locate_by_owner] called for %s\n", PROGNAME, owner);
	p = pcs[pc_ptr];
	while (p != NULL) {
		if (strcmp(p->owner, owner) == 0)
			return(p);
		p = p->nxt;
	}
	return(p);
}

struct FH *
locate_by_handle(fh)
	char *fh;
{
struct FH *p;

	if (debugmode)
		fprintf(stdout, "%s: [locate_by_handle] called\n", PROGNAME);
	p = fhs[fh_ptr];
	while (p != NULL) {
		if (memcmp(p->fh, fh, FHSIZE) == 0)
			return(p);
		p = p->nxt;
	}
	return(p);
}

int
process_lock_request(request, owner, fh, access, mode, offset, length, cookie, lockowner, exclusive)
	int request;
	char *fh, *owner;
	_uint32 access, mode, length, offset, cookie, lockowner, exclusive;
{
struct PC *pc1;
struct LK *lk1;
struct LK *lk2;
struct FH *fh1;
int i, j, len;

#define PreviousWrappedPointer(x) (x == 0)? (LOCK_CACHE_SIZE-1): (x-1)

	if (debugmode) {
	unsigned long t1=offset, t2=length, t3=cookie, t4=lockowner;
		fprintf(stdout, "%s: [process_lock_request] called with\n", PROGNAME);
		fprintf(stdout, "%s: request     = %s\n", PROGNAME, requestname(request));
		fprintf(stdout, "%s: owner       = %s\n", PROGNAME, owner);
		if (request != LOCK_REMOVE_ALL) {
			fprintf(stdout, "%s: file_handle = ", PROGNAME);
			for (j=0; j<FHSIZE; j++)
				fprintf(stdout, "%02X", fh[j] & 0xff);
			fprintf(stdout,"\n");
			if ((request == LOCK_SHARE) || (request == LOCK_UNSHARE)) {
				fprintf(stdout, "%s: access      = %s\n", PROGNAME, accessname(access));
				fprintf(stdout, "%s: mode        = %s\n", PROGNAME, modename(mode));
			}
			if ((request == LOCK_LOCK) || (request == LOCK_UNLOCK)) {
				fprintf(stdout, "%s: offset      = %lu\n", PROGNAME, t1);
				fprintf(stdout, "%s: length      = %lu\n", PROGNAME, t2);
			}
			fprintf(stdout, "%s: cookie      = %lu\n", PROGNAME, t3);
			fprintf(stdout, "%s: lockowner   = %lu\n", PROGNAME, t4);
		}
	}
	pc_ptr = OwnerHash(owner);

	if (request != LOCK_REMOVE_ALL)  {
		if ((pc1=locate_by_owner(owner)) == NULL) {
			if ((pc1=(struct PC *)malloc(sizeof(struct PC))) == NULL)
				return(FALSE);
			pc1->owner = my_strdup(owner);
			pc1->lock = NULL;
			pc1->prv = NULL;
			pc1->nxt = pcs[pc_ptr];
			for (i=0; i<LOCK_CACHE_SIZE; i++)
				pc1->cache[i].last_cookie = cookie-i-1;
			pc1->cache_pointer = 0;
			pcs[pc_ptr] = pc1;
			if ((pc1->nxt) != NULL)
				(pc1->nxt)->prv = pc1;
		}
		for (i=pc1->cache_pointer, j=0; j<LOCK_CACHE_SIZE; j++) {
			i = PreviousWrappedPointer(i);
			if (pc1->cache[i].last_cookie == cookie)
				return(pc1->cache[i].last_result);
		}
		pc1->cache[pc1->cache_pointer].last_cookie = cookie;
		if (pc1->cache_pointer == LOCK_CACHE_SIZE-1)
			pc1->cache_pointer = 0;
		else
			pc1->cache_pointer++;
	}
	switch(request) {
	case LOCK_SHARE:
		fh_ptr = fh[HASH_LOCATION] & 0xff;
		if ((fh1=locate_by_handle(fh)) == NULL) {
			if ((fh1=(struct FH *)malloc(sizeof(struct FH))) == NULL) {
				pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
				return(FALSE);
			}
			(void) memcpy(fh1->fh, fh, FHSIZE);
			fh1->lock = NULL;
			fh1->prv = NULL;
			fh1->nxt = fhs[fh_ptr];
			fhs[fh_ptr] = fh1;
			if ((fh1->nxt) != NULL)
			       (fh1->nxt)->prv = fh1;
		}
		lk1 = fh1->lock;
		while (lk1 != NULL) {
			if (lk1->type == SHARE) {
				if (!shares[mode*3+access-1][lk1->lock.s.mode*3+lk1->lock.s.access-1]) {
					lk1->num_failed++;
					pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
					return(FALSE);
				}
			}
			lk1 = lk1->nxt_fh;
		}
		if ((lk1=(struct LK *)malloc(sizeof(struct LK))) == NULL) {
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
			return(FALSE);
		}
		lk1->fh = fh1;
		lk1->pc = pc1;
		lk1->type = SHARE;
		lk1->num_failed = 0;
		lk1->lock.s.mode = mode;
		lk1->lock.s.access = access;
		lk1->prv = NULL;
		lk1->prv_fh = NULL;
		if ((lk1->nxt=pc1->lock) != NULL)
			(lk1->nxt)->prv = lk1;
		pc1->lock = lk1;
		if ((lk1->nxt_fh=fh1->lock) != NULL)
			(lk1->nxt_fh)->prv_fh = lk1;
		fh1->lock = lk1;
		pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
		return(TRUE);
	case LOCK_UNSHARE:
		fh_ptr = fh[HASH_LOCATION] & 0xff;
		if ((fh1=locate_by_handle(fh)) == NULL) {
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
			return(TRUE);
		}
		lk1 = fh1->lock;
		while (lk1 != NULL) {
			if ((lk1->pc == pc1) &&
			    (lk1->type == SHARE) &&
			    (mode == lk1->lock.s.mode) &&
			    (access == lk1->lock.s.access))
			{
				if ((lk1->prv_fh) == NULL) {
					if ((fh1->lock=lk1->nxt_fh) == NULL) {
						if (fh1->prv == NULL) {
							if ((fhs[fh_ptr]=fh1->nxt) != NULL)
								(fh1->nxt)->prv = NULL;
						}
						else {
							if (((fh1->prv)->nxt=fh1->nxt) != NULL)
								(fh1->nxt)->prv = fh1->prv;
						}
						free(fh1);
					}
					else
						(fh1->lock)->prv_fh = NULL;
				}
				else {
					if (((lk1->prv_fh)->nxt_fh=lk1->nxt_fh) != NULL)
						(lk1->nxt_fh)->prv_fh = lk1->prv_fh;
				}
				if ((lk1->prv) == NULL) {
					if ((pc1->lock=lk1->nxt) == NULL) {
						if (pc1->prv == NULL) {
							if ((pcs[pc_ptr]=pc1->nxt) != NULL)
								(pc1->nxt)->prv = NULL;
						}
						else {
							if (((pc1->prv)->nxt=pc1->nxt) != NULL)
								(pc1->nxt)->prv = pc1->prv;
						}
						free(pc1->owner);
						free(pc1);
						pc1 = NULL;
					}
					else
						(pc1->lock)->prv = NULL;
				}
				else {
					if (((lk1->prv)->nxt=lk1->nxt) != NULL)
						(lk1->nxt)->prv = lk1->prv;
				}
				free(lk1);
				if (pc1 != NULL)
					pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
				return(TRUE);
			}
			lk1 = lk1->nxt_fh;
		}
		if (pc1 != NULL)
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
		return(TRUE);
	case LOCK_LOCK:
		fh_ptr = fh[HASH_LOCATION] & 0xff;
		if ((fh1=locate_by_handle(fh)) == NULL) {
			if ((fh1=(struct FH *)malloc(sizeof(struct FH))) == NULL) {
				pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
				return(FALSE);
			}
			(void) memcpy(fh1->fh, fh, FHSIZE);
			fh1->lock = NULL;
			fh1->prv = NULL;
			fh1->nxt = fhs[fh_ptr];
			fhs[fh_ptr] = fh1;
			if ((fh1->nxt) != NULL)
				(fh1->nxt)->prv = fh1;
		}
		lk1 = fh1->lock;
		while (lk1 != NULL) {
			if (lk1->type == LOCK) {
				if (((lk1->lock.l.offset >= offset) &&
				      (offset+length > lk1->lock.l.offset)) ||
				     ((lk1->lock.l.offset < offset) &&
				      (lk1->lock.l.offset + lk1->lock.l.length > offset)))
				{
					if (lk1->lock.l.exclusive || exclusive) {
						lk1->num_failed++;
						pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
						return(FALSE);
					}
				}
			}
			lk1 = lk1->nxt_fh;
		}
		if ((lk1=(struct LK *)malloc(sizeof(struct LK))) == NULL) {
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
			return(FALSE);
		}
		lk1->fh = fh1;
		lk1->pc = pc1;
		lk1->type = LOCK;
		lk1->num_failed = 0;
		lk1->lock.l.offset = offset;
		lk1->lock.l.length = length;
		lk1->lock.l.lockowner = lockowner;
		lk1->lock.l.exclusive = exclusive;
		lk1->prv = NULL;
		lk1->prv_fh = NULL;
		if ((lk1->nxt=pc1->lock) != NULL)
			(lk1->nxt)->prv = lk1;
		pc1->lock = lk1;
		if ((lk1->nxt_fh=fh1->lock) != NULL)
			(lk1->nxt_fh)->prv_fh = lk1;
		fh1->lock = lk1;
		pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
		return(TRUE);
	case LOCK_ZAP_LOCKS:
		fh_ptr = fh[HASH_LOCATION] & 0xff;
		if ((fh1=locate_by_handle(fh)) == NULL) {
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
			return(TRUE);
		}
		lk1 = fh1->lock;
		while (lk1 != NULL) {
			if ((lk1->type != LOCK) || (lk1->lock.l.lockowner != lockowner) || (lk1->pc != pc1)) {
				lk1 = lk1->nxt;
				continue;
			}
			if ((lk1->prv_fh) == NULL) {
				if ((fh1->lock=lk1->nxt_fh) == NULL) {
					if (fh1->prv == NULL) {
						if ((fhs[fh_ptr]=fh1->nxt) != NULL)
							(fh1->nxt)->prv = NULL;
					}
					else {
						if (((fh1->prv)->nxt=fh1->nxt) != NULL)
							(fh1->nxt)->prv = fh1->prv;
					}
					free(fh1);
				}
				else
					(fh1->lock)->prv_fh = NULL;
			}
			else {
				if (((lk1->prv_fh)->nxt_fh=lk1->nxt_fh) != NULL)
					(lk1->nxt_fh)->prv_fh = lk1->prv_fh;
			}
			if ((lk1->prv) == NULL) {
				if ((pc1->lock=lk1->nxt) == NULL) {
					if (pc1->prv == NULL) {
						if ((pcs[pc_ptr]=pc1->nxt) != NULL)
							(pc1->nxt)->prv = NULL;
					}
					else {
						if (((pc1->prv)->nxt=pc1->nxt) != NULL)
							(pc1->nxt)->prv = pc1->prv;
					}
					free(pc1->owner);
					free(pc1);
					pc1 = NULL;
				}
				else
					(pc1->lock)->prv = NULL;
			}
			else {
				if (((lk1->prv)->nxt=lk1->nxt) != NULL)
					(lk1->nxt)->prv = lk1->prv;
			}
			lk2 = lk1->nxt;
			if (debugmode > 1) {
			unsigned long t1=lk1->lock.l.offset, t2=lk1->lock.l.length;
				fprintf(stdout, "%s: offset      = %lu\n", PROGNAME, t1);
				fprintf(stdout, "%s: length      = %lu\n", PROGNAME, t2);
			}
			free(lk1);
			lk1 = lk2;
		}
		if (pc1 != NULL)
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
		return(TRUE);
	case LOCK_REMOVE_ALL:
		if ((pc1=locate_by_owner(owner)) == NULL)
			return(TRUE);
		if (pc1->prv == NULL) {
			if ((pcs[pc_ptr]=pc1->nxt) != NULL)
				(pc1->nxt)->prv = NULL;
		}
		else {
			if (((pc1->prv)->nxt=pc1->nxt) != NULL)
				(pc1->nxt)->prv = pc1->prv;
		}
		lk1 = pc1->lock;
		while (lk1 != NULL) {
			if ((lk1->prv_fh) == NULL) {
				fh1 = lk1->fh;
				if ((fh1->lock=lk1->nxt_fh) == NULL) {
					if (fh1->prv == NULL) {
						fh_ptr = fh1->fh[HASH_LOCATION] & 0xff;
						if ((fhs[fh_ptr]=fh1->nxt) != NULL)
							(fh1->nxt)->prv = NULL;
					}
					else {
						if (((fh1->prv)->nxt=fh1->nxt) != NULL)
							(fh1->nxt)->prv = fh1->prv;
					}
					free(fh1);
				}
				else
					(fh1->lock)->prv_fh = NULL;
			}
			else
				if (((lk1->prv_fh)->nxt_fh=lk1->nxt_fh) != NULL)
					(lk1->nxt_fh)->prv_fh = lk1->prv_fh;
			lk2 = lk1->nxt;
			if (debugmode > 1) {
			unsigned long t1=lk1->lock.l.offset, t2=lk1->lock.l.length;
				fprintf(stdout, "%s: offset      = %lu\n", PROGNAME, t1);
				fprintf(stdout, "%s: length      = %lu\n", PROGNAME, t2);
			}
			free(lk1);
			lk1 = lk2;
		}
		free(pc1->owner);
		free(pc1);
		return(TRUE);
	case LOCK_UNLOCK:
		fh_ptr = fh[HASH_LOCATION] & 0xff;
		if ((fh1=locate_by_handle(fh)) == NULL) {
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
			return(TRUE);
		}
		lk1 = fh1->lock;
		while (lk1 != NULL) {
			if ((lk1->pc == pc1) &&
			    (lk1->type == LOCK) &&
			    (offset == lk1->lock.l.offset) &&
			    (length == lk1->lock.l.length))
			{
				if ((lk1->prv_fh) == NULL) {
					if ((fh1->lock=lk1->nxt_fh) == NULL) {
						if (fh1->prv == NULL) {
							if ((fhs[fh_ptr]=fh1->nxt) != NULL)
								(fh1->nxt)->prv = NULL;
						}
						else {
							if (((fh1->prv)->nxt=fh1->nxt) != NULL)
								(fh1->nxt)->prv = fh1->prv;
						}
						free(fh1);
					}
					else
						(fh1->lock)->prv_fh = NULL;
				}
				else {
					if (((lk1->prv_fh)->nxt_fh=lk1->nxt_fh) != NULL)
						(lk1->nxt_fh)->prv_fh = lk1->prv_fh;
				}
				if ((lk1->prv) == NULL) {
					if ((pc1->lock=lk1->nxt) == NULL) {
						if (pc1->prv == NULL) {
							if ((pcs[pc_ptr]=pc1->nxt) != NULL)
								(pc1->nxt)->prv = NULL;
						}
						else {
							if (((pc1->prv)->nxt=pc1->nxt) != NULL)
								(pc1->nxt)->prv = pc1->prv;
						}
						free(pc1->owner);
						free(pc1);
						pc1 = NULL;
					}
					else
						(pc1->lock)->prv = NULL;
				}
				else {
					if (((lk1->prv)->nxt=lk1->nxt) != NULL)
						(lk1->nxt)->prv = lk1->prv;
				}
				free(lk1);
				if (pc1 != NULL)
					pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
				return(TRUE);
			}
			lk1 = lk1->nxt_fh;
		}
		if (pc1 != NULL)
			pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = TRUE;
		return(TRUE);
	}
	if (pc1 != NULL)
		pc1->cache[PreviousWrappedPointer(pc1->cache_pointer)].last_result = FALSE;
	return(FALSE);
}
