/* $Header: /s/misc/Program
Files/Hummingbird/Connectivity/10.00/NFSClient/utility/hclnfsd.c 31/06/04 1:28p Frisch $ */
#define VERSION  "10.0.0.2"
/*
HCLNFSD (formerly BWNFSD) is an RPC daemon used in conjunction with
HCLNFS (a client NFS for DOS based PCs). HCLNFSD provides Authentication,
Print Spooling, DOS 3.1 Locking, DOS 3.1 Sharing, GID name mapping,
and UID name mapping services for HCLNFS and associated applications on
the PC. HCLNFSD is also used by some Macintosh NFS clients.

The HCLNFSD code is originally copyright Beame & Whiteside Software Ltd.
and now is released to the Public Domain. The intent is that all server
vendors include a version of HCLNFSD with their operating system. HCLNFSD
can run simultanteously with PCNFSD, but provides more features.

	Please send modifications to:

		Hummingbird Ltd.
		1 Sparks Avenue
		North York, Ontario
		Canada	M2H 2W1
		+1 (416) 496-2200
		support@hummingbird.com

Please modify by including "ifdefs" for your particular operating
system, with appropriate modifications to the "makefile".

MODIFICATION HISTORY
--------------------
11/11/91 gsw Added some debug statements
24/03/92 gsw Changed SHADOW to SHADW since it conflicted on some SYS-V's
31/07/92 fjw Changed printers code to check for abnormal conditions
31/07/92 fjw Cleaned up the code and repaired the printer parsing
01/08/92 fjw Merged in SVR4 code from Frank Glass (gatech!holos0!fsg)
31/08/92 cfb Major changes to support pcnfsdv2-style queues, etc.
31/08/92 fjw Changed all the ifdefs around for cleaner code
31/08/92 fjw Added user cacheing in authorize code
08/10/92 fjw Fixed code relating to printing and debugmode
05/11/92 fjw Added some pretty print in debug mode
14/05/93 fjw Changed the daemon-detach code to be more like Stevens book
14/05/93 fjw Consolidated #defines and added support for Univel and Solaris
09/02/94 fjw Changed most of the manifest constant names to assist customisation
09/02/94 fjw Added code to allow locks/file/process to be removed
08/05/94 fjw Fixed set_auth_parameters() *again* (code got lost somehow)
08/23/94 cpz Fixed possible undefined SIGCLD; added conditional for netinet
08/23/94 cpz Fix debug write in get_gr_gid; Carl Zwanzig
09/26/94 dp  Change to the makefile for BSDI
11/25/94 fjw Add checks for xdr_string overruns, thanks to Paul Freeman
12/09/94 fjw Add support for Linux (SYSV_PGRP), thanks to Doug Hester
06/05/95 fjw Fix returned UID as a number if name not found
06/05/95 fjw Fix debug code printing of gids for uid, thanks to Denny
06/23/95 fjw Put in workaround for Solaris pstatus files, thanks B. Beckberger
07/24/95 fjw Rewrite to enforce 32-bit types for OSF/1; molto changes in source
08/09/95 fjw Cast all the pointers to the svc_routines to make UnixWare happy
09/13/95 fjw Change code to allow program name to be user-selectable
10/13/95 fjw Change uint32 to _uint32 to keep SCO happy, thanks to Kevin Inscoe
11/01/95 fjw Fix editing error, thanks to AMV Sweden and Lars Hornberg
11/15/95 dp  Added Solaris 2.3 option to the HCLNFSD daemon
12/12/95 su  Added DEC OSF/1 v3.0 w/protected passwd dbase, thanks Stan Uno
12/14/95 fjw Allow printqueues whose names are longer than 11 characters
12/15/95 fjw Make changes to processing of child processes
01/12/95 smd Added daemonized loggin option ( -l )
01/12/95 smd Added logging of uid,euid,gid and egid when starting daemon log
05/13/96 fjw Fixed the Solaris2.5 open-file-from-name-routines bug
09/03/96 fjw Added a little error-checking code and some new systems
11/26/96 fjw replaced the strdup's i missed, and added ultrix 4.3 to makefile
03/22/96 fjw added SIGUSR2 processing to dump lofile a little more nicely
04/01/97 cfb added exclusive and read shared locking
04/10/97 fjw added a timeout to the uid/gid cache
07/23/97 fjw added code for NeXT, thanks to hml@llnl.gov
04/13/98 fjw added defines to remove username/password caching
07/06/98 fjw fixed Unixware problem, thanks Terry.Watt@TheAA.com
11/27/98 fjw added code to help printfile name collisions
03/25/99 fjw added support for HPUX11 (cast routines to xdrproc_t)
04/23/99 mcf changed pointer cast
06/25/99 fjw fixed bug in locking code
04/26/05 mcf added support for binding to user-specified port
*/

#define USAGE "usage: %s [-A] [-d [-d] | -l log] [-X] [-s file] [staging_directory] [-p port]\n"

#define PROGNAME "hclnfsd"

#define MAX_NUM_PRINTERS 500
#define POLL_INTERVAL 60
#define MAX_NUM_CANCEL_CACHE 10
#define NUM_USER_CACHE 3
#define UID_GID_CACHE_TIMEOUT	(60*60*24)    /* 1 day of seconds */

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>

#ifdef FIND_QUEUES_AIX
#include <sys/select.h>
#endif

#include <sys/stat.h>
#include <rpc/rpc.h>

#ifndef INCLUDE_BOGOSITY
#include <rpc/clnt.h>	/* added for svr4 , usually included by <rpc/rpc.h> */
#include <rpc/svc.h>
#endif

#include <errno.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <grp.h>
#include <string.h>

#ifndef DONT_USE_NETINET_IN
#include <netinet/in.h>   /* This is problematic on some systems, required on others */
#endif

#ifdef USE_BCOPY
#define memcmp(a,b,c) bcmp((b),(a),(c))
#define memcpy(a,b,c) bcopy((b),(a),(c))
#else
#include <memory.h>
#endif

#ifdef NeXT
#define O_NOCTTY	0x8000
#define S_IRUSR 	0000400 /* read permission, owner */
#define S_IWUSR 	0000200 /* write permission, owner */
#define S_IRGRP 	0000040 /* read permission, group */
#define S_IROTH 	0000004 /* read permission, other */
#endif

#ifdef SHADW
#include <shadow.h>
#endif

#ifndef INT32_LONG
typedef unsigned int _uint32;
#else
typedef unsigned long _uint32;
#endif

#ifdef SGI
extern struct group *getgrent();
#endif

char *crypt();
char *my_strdup();
void free_printer();
void get_printers();

#ifdef INT_SIG
#define Void int
#define Return return(0)
#else
#define Void void
#define Return return
#endif

#define Generic_t void *

#ifdef WINNFS
#include <tiuser.h>
#endif

#ifdef FIND_QUEUES_SVR4
#define ALLOW 0
#define DENY  1
#define CLASS 2

struct {
	int access;
	int count;
	union {
		char **users;
		int  *printers;
	} n;
} printer_access[MAX_NUM_PRINTERS];
#endif

#define BW_NFS_PROGRAM 0x2f00dbadL
#define BW_NFS_VERSION 1L

#define SPOOL_INQUIRE  1L	/* Return fhandle of spool directory */
#define SPOOL_FILE     2L	/* Spool file */
#define AUTHORIZE      3L	/* Authorize and return UID and GIDs (max 16) */
#define GRP_NAME_TO_NUMB 4L	/* Convert group name to number */
#define GRP_TO_NUMBER  5L	/* Convert group number(s) to name(s) */
#define RETURN_HOST    6L	/* Convert IP to hostname */
#define UID_TO_NAME    7L	/* Convert UID(s) to name(s) */
#define NAME_TO_UID    8L	/* Convert name(s) to UID(s)  L. Yen */
#define SHARE	      20L	/* DOS 3.1 Share function */
#define UNSHARE       21L	/* DOS 3.1 UnShare function */
#define LOCK	      22L	/* DOS 3.1 Lock request */
#define REMOVE	      23L	/* Remove all locks/shares for PC */
#define UNLOCK	      24L	/* DOS 3.1 Unlock request */
#define GET_PRINTERS  30L	/* Get a list of printers */
#define GET_PRINTQ    31L	/* Get print queue entries and status */
#define CANCEL_PRJOB  32L	/* Cancel a print job */
#define ZAP_LOCKS    105L	/* Remove locks on file by process */

#define NUSRS  10		/* Maximum names to convert */

#define WNPRQ_ACTIVE	0
#define WNPRQ_PAUSE	1
#define WNPRQ_ERROR	2
#define WNPRQ_PENDING	3
#define WNPRQ_PROBLEM	4
#define WN_SUCCESS	0
#define WN_BAD_QUEUE	0x43

static char dir[32], g_string[32], spool_dir[132], file[256], file1[256], *p;
static char remote_host[16], printer[65], *jobname, file_name[9], *hostname;
static int num_printers, extension, lgids[NGRPS+1], luids[NUSRS+1];
static int A_flag=0, s_flag=0, do_print;
static _uint32 my_ip;
static time_t GidCacheTimeoutTime=0;
int debugmode, DontDoPrinting;
char *debuglog;

struct queueentry {
	int jobid;
	char *username;
	char *params;
	int queueposition;
	int jobstatus;
	unsigned long timesubmitted;
	unsigned long size;
	int copies;
	char *jobcomment;
	struct queueentry *next;
};

struct printertype {
	char *queuename;
	char *queuecomment;
	int queuestatus;
	int numphysicalprinters;
	unsigned long lasttime;
	struct queueentry *head;
} printers[MAX_NUM_PRINTERS];

static struct mm_ips {		/* Valid hosts to authenticate for */
	char   type;
	_uint32 ip;
	_uint32 mask;
	_uint32 redirect_ip;
} **ips;

static struct mm_uids { 	/* Cache of uid to name translation */
	int   mm_uid;
	char *mm_username;
} **uds;

static struct mm_gidss {	/* Cache of uid and associated gids */
	int mm_uid;
	int mm_gid;
	int mm_gid_count;
	int mm_gids[NGRPS];
} **gds;

static int gids_count=0, uids_count=0;
static int ips_count=0;

static struct locking {
	_uint32 lockowner;	/* 4 bytes */
	_uint32 cookie; 	/* 4 bytes */
	char   name[17];       /* 16 character name */
	char   fh[32];	       /* 32 bytes */
	_uint32 mode;
	_uint32 access;
	_uint32 offset;
	_uint32 length;
	_uint32 stat;
	_uint32 sequence;
	_uint32 exclusive;
} lock = { 0 };

static struct printqs {
	char queuename[32];
	char username[32];
	int printer_number;
} printq;

static struct cancaches {
	char queuename[32];
	char username[32];
	int jobid;
	int response;
} cancache[MAX_NUM_CANCEL_CACHE] = { 0 };

int current_cache=0;

int
my_strnicmp(p1, p2, num)
	char *p1, *p2;
	int num;
{
char c, d;

	for ( ; num>0; num--) {
		c = *p1++;
		d = *p2++;
		c = islower(c)? toupper(c): c;
		d = islower(d)? toupper(d): d;
		if (c != d)
			return(1);
	}
	return(0);
}

#ifdef CHECK_PRINTER_ACCESS
int
check_access_printer(pn, name)
	int pn;
	char *name;
{
int i;
char **p;

	if (debugmode)
		fprintf(stdout, "%s: [check_access_printer] pn = <%d>, name = <%s>\n", PROGNAME, pn, name);

	if (printer_access[pn].count == -1) {
		if (debugmode)
			fprintf(stdout, "%s: [check_access_printer] deny all requests\n", PROGNAME);
		return(0);				     /* Deny all */
	}
	p = printer_access[pn].n.users;
	for (i=0; i<printer_access[pn].count; i++) {
		if (strcmp(*p++, name) == 0) {
			if (debugmode)
				fprintf(stdout, "%s: [check_access_printer] request %s\n", PROGNAME, (printer_access[pn].access==ALLOW)? "allowed": "denied");
			return(printer_access[pn].access == ALLOW);
		}
	}
	if (debugmode)
		fprintf(stdout, "%s: [check_access_printer] request %s\n", PROGNAME, (printer_access[pn].access==DENY)? "allowed": "denied");
	return(printer_access[pn].access == DENY);
}

int
check_access(pn, name)
	int pn;
	char *name;
{
int i;
int *pi;

	if (debugmode)
		fprintf(stdout, "%s: [check_access] pn = <%d>, name = <%s>\n", PROGNAME, pn, name);

	if (printer_access[pn].access != CLASS)
		return(check_access_printer(pn, name));
	pi = printer_access[pn].n.printers;
	for (i=0; i<printer_access[pn].count; i++)
		if (check_access_printer(*pi++, name)) {
			if (debugmode)
				fprintf(stdout, "%s: [check_access] returns Allow\n", PROGNAME);
			return(1);
		}
	if (debugmode)
		fprintf(stdout, "%s: [check_access] returns Deny\n", PROGNAME);
	return(0);
}
#endif

void
clear_uid_gid_cache()
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [clear_uid_gid_cache]\n", PROGNAME);
	for (i=0; i<uids_count; i++) {
		free((*(uds+i))->mm_username);
		free(*(uds+i));
	}
	if (uids_count > 0) {
		free(uds);
		uids_count = 0;
	}
	endpwent();
	for (i=0; i<gids_count; i++)
		free(*(gds+i));
	if (gids_count > 0) {
		free(gds);
		gids_count = 0;
	}
	endgrent();
	GidCacheTimeoutTime = time(NULL) + UID_GID_CACHE_TIMEOUT;
}

Void
clear(unused)
	int unused;
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [clear]\n", PROGNAME);
	clear_uid_gid_cache();
	for (i=0; i<num_printers; i++) {
		free_printer(i);
		free(printers[i].queuename);
		if (printers[i].queuecomment != NULL)
			free(printers[i].queuecomment);
	}
	get_printers();
	(void) signal(SIGHUP, clear);
	Return;
}

Void
free_child(unused)
	int unused;
{
	if (debugmode)
		fprintf(stdout, "%s: [free_child] called\n", PROGNAME);

/*
 * Make this wait for _any_ outstanding children ...
 */
	while (waitpid(-1, NULL, WNOHANG) > 0)
		;
#ifdef SIGCHLD
	(void) signal(SIGCHLD, free_child);
#endif
#ifdef SIGCLD
	(void) signal(SIGCLD, free_child);
#endif
	Return;
}

void
strlwr(p)
	char *p;
{
	do {
		if ((*p >= 'A') && (*p <= 'Z'))
			*p += 'a' - 'A';
	} while (*(++p));
}

void
getmask(ip, mask)
	_uint32 *ip, *mask;
{
	if (*ip == (_uint32)0L)
		*mask = (_uint32)0L;
	else if (*ip <= (_uint32)0xffL) {
		*ip <<= 24;
		*mask = (_uint32)0xff000000L;
	}
	else if (*ip <= (_uint32)0xffffL) {
		*ip <<= 16;
		*mask = (_uint32)0xffff0000L;
	}
	else if (*ip <= (_uint32)0xffffffL) {
		*ip <<= 8;
		*mask = (_uint32)0xffffff00L;
	}
	else
		*mask = (_uint32)0xffffffffL;
}

_uint32
my_inet_addr(cp)
	char *cp;
{
static union {
	 _uint32 inn;
	 struct {
		 unsigned char b1, b2, b3, b4;
	 } i;
} in;
int b1, b2, b3, b4;

	b1 = b2 = b3 = b4 = -1;
	if (sscanf(cp, "%d.%d.%d.%d", &b1, &b2, &b3, &b4) == -1)
		return((_uint32)~0L);
	if (b4 == -1) {
		b4 = b3;
		b3 = b2;
		b2 = 0;
	}
	if (b4 == -1) {
		b4 = b3;
		b3 = 0;
	}
	if (((unsigned int) b1 > 255) || ((unsigned int) b2 > 255) ||
	    ((unsigned int) b3 > 255) || ((unsigned int) b4 > 255))
		return((_uint32)~0L);
	in.i.b1 = (unsigned char)b1;
	in.i.b2 = (unsigned char)b2;
	in.i.b3 = (unsigned char)b3;
	in.i.b4 = (unsigned char)b4;
	return((_uint32)ntohl(in.inn));
}

/*
 *  For -s parameter, returns IP addresses and the associated mask to determine
 *  if the IP address of the destination machine/network is matches an entry
 *  in the -s file.
 */
int
convert_ip(s, ip, mask)
	char *s;
	_uint32 *ip, *mask;
{
_uint32 ia, lip, lmask;
struct hostent *he;
struct netent *ne;

	if (debugmode)
		fprintf(stdout, "%s: [convert_ip] s = <%s>\n", PROGNAME, s);

	if ((ia=inet_network(s)) != (_uint32)~0L) {
		lip = ia;
		getmask(&lip, &lmask);
	}
	else if ((ia=my_inet_addr(s)) != (_uint32)~0L) {
		lip = ia;
		lmask = (_uint32)0xffffffffL;
	}
	else if ((ne=getnetbyname(s)) != NULL) {
		lip = ne->n_net;
		getmask(&lip, &lmask);
	}
	else if ((he=gethostbyname(s)) != NULL) {
		lip = *(_uint32 *)he->h_addr;
		lmask = (_uint32)0xffffffffL;
	}
	else {
		if (debugmode)
			fprintf(stdout, "%s: [convert_ip] unable to convert <%s> into an IP address\n", PROGNAME, s);
		return(0);
	}
	*ip = lip;
	if (mask != NULL)
		*mask = lmask;
	if (debugmode) {
	unsigned long t1=lip, t2=lmask;
		fprintf(stdout, "%s: [convert_ip] Success: ip = <%l08X>, mask = <%08lX>\n", PROGNAME, t1, t2);
	}
	return(1);
}

/*
 *	 Read the -s file which lists rules for which hosts can be
 *	 authenticated. Each line is converted to a IP address and mask. Any
 *	 incoming address is ANDed with the mask and compared to IP address,
 *	 if it matches, the action associated with the line is taken.
 *
 *	 Valid actions:
 *		 "+" - Perform Validatation check
 *		 "-" - Return Authorization error
 *		 "=" - Tell client to attempt Authorization at new address
 */
void
read_ips(file)
	char *file;
{
FILE *f;
char s1[255], s2[255], line[255], type;
int i;
_uint32 ip, mask, redirect_ip;

	if (debugmode)
		fprintf(stdout, "%s: [read_ips] file = %s\n", PROGNAME, file);

	if ((f=fopen(file,"r")) == NULL) {
		fprintf(stdout, "%s: [read_ips] file \"%s\" not found.\n", PROGNAME, file);
		fflush(stdout);
		exit(1);
	}

	ips = (struct mm_ips **)malloc(1);
	while (fgets(line, sizeof(line)-1, f) != NULL) {
		if ((type=line[0]) != '#') {
			i = sscanf(&line[1],"%s %s", s1, s2);
			if (debugmode)
				fprintf(stdout, "%s: [read_ips] parsing <%s>\n", PROGNAME, &line[1]);

			if ((i == EOF) || (i == 0) || (i > 2) || ((type != '+') &&
			    (type != '-') && (type != '=')) || ((type == '=') &&
			    (i != 2)) || ((type != '=') && (i != 1))) {
				fprintf(stdout, "%s: [read_ips] syntax error, line ignored : %s", PROGNAME, line);
			}
			else {
				if (!convert_ip(s1, &ip, &mask)) {
					fprintf(stdout, "%s: [read_ips] \"%s\" invalid address.\n", PROGNAME, s1);
					continue;
				}
				if (i == 2) {
					if (!convert_ip(s2, &redirect_ip, NULL)) {
					       fprintf(stdout, "%s: [read_ips] \"%s\" invalid address.\n", PROGNAME, s1);
					       continue;
					}
				}
				ips = (struct mm_ips **)realloc(ips,sizeof(ips)*(ips_count+1));
				*(ips+ips_count) = (struct mm_ips *)malloc(sizeof(struct mm_ips));
				(*(ips+ips_count))->type = type;
				(*(ips+ips_count))->ip = ip;
				(*(ips+ips_count))->mask = mask;
				(*(ips+ips_count))->redirect_ip = redirect_ip;
				ips_count++;
		       }
		}
	}
	fclose(f);
	if (debugmode)
		fprintf(stdout, "%s: [read_ips] exiting with %u values\n", PROGNAME, ips_count);
}

/*
 *	 Add UID to USERNAME translation to cache
 */
void
add_uid(uid, username)
	int uid;
	char *username;
{
	if (debugmode)
		fprintf(stdout, "%s: [add_uid] uid <%d>, name <%s>\n", PROGNAME, uid, username);
	if (uids_count++ == 0)
		uds = (struct mm_uids **)malloc(1);

	uds = (struct mm_uids **)realloc(uds, sizeof(uds)*uids_count);
	(*(uds+uids_count-1)) = (struct mm_uids *)malloc(sizeof(struct mm_uids));
	(*(uds+uids_count-1))->mm_uid = uid;
	(*(uds+uids_count-1))->mm_username = (char *)malloc(strlen(username)+1);
	(void)strcpy((*(uds+uids_count-1))->mm_username, username);
}

/*
 *	 Lookup USERNAME given a UID, also add translation to cache
 */
char *
get_ui_name(uid)
	int uid;
{
struct passwd *pww;
int i;
static char ui[32];

	if (debugmode)
		fprintf(stdout, "%s: [get_ui_name] uid <%u>\n", PROGNAME, uid);

	if (time(NULL) > GidCacheTimeoutTime)
		clear_uid_gid_cache();
	for (i=0; i<uids_count; i++)
		if ((*(uds+i))->mm_uid == uid)
			return((*(uds+i))->mm_username);
	endpwent();
	if ((pww=getpwuid(uid)) == NULL) {
		sprintf(ui,"%d", uid);
		if (debugmode)
			fprintf(stdout, "%s: [get_ui_name] getpwuid returned <%s>\n", PROGNAME, ui);
		return(ui);
	}
	add_uid(uid, pww->pw_name);
	if (debugmode)
		fprintf(stdout, "%s: [get_ui_name] returns <%s>\n", PROGNAME, pww->pw_name);
	return(pww->pw_name);
}

/*
 *	 Given a UID, GID and USERNAME, create a list of GIDs associated
 *	 with the UID, a maximum of NGRPS GIDS are used.
 */
void
fill_gid(gd, uid, gid, username)
	struct mm_gidss *gd;
	int uid, gid;
	char *username;
{
struct group *gr, *getgrent();
char **p;
int flag;

	add_uid(uid, username);
	if (debugmode)
		fprintf(stdout, "%s: [fill_gid] uid = %u, gid = %u, user = %s\n", PROGNAME, uid, gid, username);
	flag = 0;
	gd->mm_uid = uid;
	gd->mm_gid = gid;
	gd->mm_gid_count = 0;
	(void) setgrent();
	while ((gr=getgrent()) != NULL) {
		p = gr->gr_mem;
		while (*p && (gd->mm_gid_count < NGRPS)) {
			if (strcmp(*p, username) == 0) {
				gd->mm_gids[gd->mm_gid_count++] = gr->gr_gid;
				if (gr->gr_gid == gid)
					flag = 1;
				break;
			}
			p++;
		}
	}
	if (!flag && (gd->mm_gid_count < NGRPS))
		gd->mm_gids[gd->mm_gid_count++] = gid;
	(void) endgrent();
}

/*
 *	 Return a list of GIDs for a given UID. The cache of GIDS/UID is
 *	 updated and used.
 */
struct mm_gidss *
get_gids(uid, gid, username)
	int uid, gid;
	char *username;
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [mm_gidss] uid = %u, gid = %u, name = %s\n", PROGNAME, uid, gid, username);

	if (time(NULL) > GidCacheTimeoutTime)
		clear_uid_gid_cache();
	for (i=0; i<gids_count; i++)
		if ((*(gds+i))->mm_uid == uid) {
			if (debugmode)
				fprintf(stdout, "%s: [mm_gidss] gids_count = %u\n", PROGNAME, gids_count);
			return(*(gds+i));
		}

	if (gids_count++ == 0)
		gds = (struct mm_gidss **)malloc(1);

	gds = (struct mm_gidss **)realloc(gds, sizeof(gds)*gids_count);
	(*(gds+gids_count-1)) = (struct mm_gidss *)malloc(sizeof(struct mm_gidss));
	(void) fill_gid((*(gds+gids_count-1)), uid, gid, username);
	if (debugmode)
		fprintf(stdout, "%s: [mm_gidss] gids_count = %u\n", PROGNAME, gids_count);
	return(*(gds+gids_count-1));
}

static struct my_groups {
	int my_gid;
	char *my_name;
} **g;
static int group_count=0;

/*
 *	 Return the text name of a GID, if none is found, then the GID
 *	 is convert to a text number.
 */
char *
get_gr_name(gid)
	int gid;
{
static char gd[32];
int i;

	if (debugmode)
		fprintf(stdout, "%s: [get_gr_name] for gid %u\n", PROGNAME, gid);

	for (i=0; i<group_count; i++) {
		if ((*(g+i))->my_gid == gid) {
			if (debugmode)
				fprintf(stdout, "%s: [get_gr_name] returns <%s>\n", PROGNAME, ((*(g+i))->my_name));
			return ((*(g+i))->my_name);
		}
	}
	sprintf(gd, "%d", gid);
	if (debugmode)
		fprintf(stdout, "%s: [get_gr_name] returns <%s>\n", PROGNAME, gd);
	return(gd);
}

/*
 *	 Given a group name, return the GID associated with it.
 */
int *
get_gr_gid(name)
	char *name;
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [get_gr_gid] for <%s>\n", PROGNAME, name);
	for (i=0; i<group_count; i++)
		if (strcmp((*(g+i))->my_name,name) == 0) {
			if (debugmode)
				fprintf(stdout, "%s: [get_gr_gid] returns %d\n", PROGNAME, (*(g+i))->my_gid);
			return(&(*(g+i))->my_gid);
		}
	if (debugmode)
		fprintf(stdout, "%s: [get_gr_gid] No Gid Found for <%s>; return failure\n", PROGNAME, name);
	return(NULL);
}

/*
 *	 Cache a list of all group names and their associated GIDS.
 */
void
get_groups()
{
struct group *gr, *getgrent();

	if (debugmode)
		fprintf(stdout, "%s: [get_groups] called\n", PROGNAME);

	g = (struct my_groups **)malloc(1);
	(void) setgrent();
	while ((gr=getgrent()) != NULL) {
		group_count++;
		g = (struct my_groups **)realloc(g, sizeof(g)*group_count);
		(*(g+group_count-1)) = (struct my_groups *)malloc(sizeof(struct my_groups));
		(*(g+group_count-1))->my_name = (char *)malloc(strlen(gr->gr_name)+1);

		(void) strcpy((*(g+group_count-1))->my_name, gr->gr_name);
		(*(g+group_count-1))->my_gid = gr->gr_gid;
	}
	(void) endgrent();
}

#ifdef FIND_QUEUES_SVR4
void
get_class_list(name, pn)
	char *name;
	int pn;
{
FILE *f;
char line[128], file_name[64];
int i;

	if (debugmode)
		fprintf(stdout, "%s: [get_class_list] name = <%s>, pn = %u\n", PROGNAME, name, pn);
	printer_access[pn].access = CLASS;
	printer_access[pn].n.printers = NULL;
	printer_access[pn].count = 0;
	sprintf(file_name, "/usr/spool/lp/admins/lp/classes/%s", name);

	if ((f=fopen(file_name,"r")) == NULL) {
		if (debugmode)
			fprintf(stdout, "%s: [get_class_list] open file %s returned NotFound\n", PROGNAME, file_name);
		return;
	}
	while (fgets(line, sizeof(line), f) != NULL) {
		*(line+strlen(line)-1) = '\0';
		for (i=0; i<pn; i++) {
			if (strcmp(line,printers[i].queuename) == 0) {
				if (printer_access[pn].count++ == 0)
					printer_access[pn].n.printers = (int *)malloc(sizeof(int));
				else
					printer_access[pn].n.printers = (int *)realloc(printer_access[pn].n.printers, printer_access[pn].count * sizeof(int));
				*(printer_access[pn].n.printers+printer_access[pn].count-1) = i;
				break;
			}
		}
	}
	fclose(f);
}

void
get_access(name, pn)
	char *name;
	int pn;
{
FILE *f;
char line[128], file_name[64], **p;
int i;

	if (debugmode)
		fprintf(stdout, "%s: [get_access] name = <%s>, pn = %u\n", PROGNAME, name, pn);
	printer_access[pn].access = ALLOW;
	printer_access[pn].n.users = NULL;
	printer_access[pn].count = 0;
	sprintf(file_name, "/usr/spool/lp/admins/lp/printers/%s/users.allow", name);
	if ((f=fopen(file_name,"r")) != NULL) {
		if (debugmode)
			fprintf(stdout, "%s: [get_access] reading %s\n", PROGNAME, file_name);
		while (fgets(line, sizeof(line), f) != NULL) {
			*(line+strlen(line)-1) = '\0';
			if ((strcmp(line, "any") == 0) ||
			    (strcmp(line, "all") == 0) ||
			    (strcmp(line, "all!all") == 0)) {
				printer_access[pn].n.users = NULL;
				printer_access[pn].count = 0;
				break;
			}
			if (printer_access[pn].count++ == 0)
				printer_access[pn].n.users = (char **)malloc(sizeof(char**));
			else
				printer_access[pn].n.users = (char **)realloc(printer_access[pn].n.users,printer_access[pn].count * sizeof(char**));
			*(printer_access[pn].n.users+printer_access[pn].count-1) = my_strdup(line);
		}
		fclose(f);
		if (printer_access[pn].count != 0) {
			if (debugmode) {
				fprintf(stdout, "%s: [get_access] access count = %u\n%s: [get_access] Allowed users:\n", PROGNAME, printer_access[pn].count, PROGNAME);
				p = printer_access[pn].n.users;
				for (i=0; i<printer_access[pn].count; i++)
					fprintf(stdout, "%s: [get_access]               %s\n", PROGNAME, *p++);
			}
			return;
		}
	}
	else {
		if (debugmode)
			fprintf(stdout, "%s: [get_access] unable to open %s (not a problem)\n", PROGNAME, file_name);
	}
	printer_access[pn].access = DENY;
	sprintf(file_name, "/usr/spool/lp/admins/lp/printers/%s/users.deny", name);
	if ((f=fopen(file_name,"r")) != NULL) {
		while (fgets(line, sizeof(line), f) != NULL) {
			*(line+strlen(line)-1) = '\0';
			if ((strcmp(line, "any") == 0) ||
			    (strcmp(line, "all") == 0)) {
				printer_access[pn].n.users = NULL;
				printer_access[pn].count = -1;
				break;
			}
			if (printer_access[pn].count++ == 0)
				printer_access[pn].n.users = (char **)malloc(sizeof(char**));
			else
				printer_access[pn].n.users = (char **)realloc(printer_access[pn].n.users,printer_access[pn].count * sizeof(char**));
			*(printer_access[pn].n.users+printer_access[pn].count-1) = my_strdup(line);
		}
		fclose(f);
		if (printer_access[pn].count != 0) {
			if (debugmode) {
				fprintf(stdout, "%s: [get_access] access count = %u\n%s: [get_access] Denied users:\n", PROGNAME, printer_access[pn].count, PROGNAME);
				p = printer_access[pn].n.users;
				for (i=0; i<printer_access[pn].count; i++)
					fprintf(stdout, "%s: [get_access]               %s\n", PROGNAME, *p++);
			}
		}
	}
	else {
		if (debugmode)
			fprintf(stdout, "%s: [get_access] unable to open %s (not a problem)\n", PROGNAME, file_name);
	}
}

void
get_printer_list(status_file, type)
	char *status_file;
	int type;
{
FILE *f;
char line[128], name[32];
int  i;

	if (debugmode)
		fprintf(stdout, "%s: [get_printer_list] file = %s, type = %u\n", PROGNAME, status_file, type);
	if ((f=fopen(status_file, "r")) == NULL) {
		if (debugmode)
			fprintf(stdout, "%s: [get_printer_list] unable to open %s\n", PROGNAME, status_file);
		return;
	}
	while (fgets(line, sizeof(line), f) != NULL)
		if ((strncmp(line, "=====", 5) == 0) ||
		    (strncmp(line, "+=====", 6) == 0)) { /* for Solaris */
			if (fgets(name, sizeof(name), f) == NULL)
				break;
			*(name+strlen(name)-1) = '\0';
			if (fgets(line, sizeof(line), f) == NULL)
				break;
			*(line+strlen(line)-1) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [get_printer_list] name = %s, line = %s\n", PROGNAME, name, line);
			for (i=0; i<(int)strlen(line); i++)
				if (strncmp(&line[i], "accepting", 9) == 0)
					break;
			if (i == strlen(line))
				continue;
			if ((strlen(name) > 11) && debugmode)
				fprintf(stdout, "%s: [get_printer_list] Printer name \"%s\" exceeds 11 characters; DOS client may not link\n", PROGNAME, name);
			if (num_printers >= MAX_NUM_PRINTERS) {
				if (debugmode)
					fprintf(stdout, "%s: [get_printer_list] Max number of printers (%d) exceeded.  Recompile with larger value\n", PROGNAME, MAX_NUM_PRINTERS);
				break;
			}
			printers[num_printers].queuename = my_strdup(name);
			printers[num_printers].queuecomment = NULL;
			if (type == 0)
				get_access(name, num_printers++);
			else
				get_class_list(name, num_printers++);
		}
	fclose(f);
}
#endif




#ifdef FIND_QUEUES_SOLARIS26
void
get_printer_list(status_file)
	char *status_file;
{
	FILE 
		*f;	
		
	char
		comment[128],
		line[128], 
		name[32], 
		printerLine[1024],
		*charPtr,
		*charPtr1;
	
	int  
		i,
		sameLine;

	if (debugmode)
		fprintf(stdout, "%s: [get_printer_list] file = %s\n", PROGNAME, status_file);

	if ((f=fopen(status_file, "r")) == NULL) 
	{
		if (debugmode)
			fprintf(stdout, "%s: [get_printer_list] unable to open %s\n", PROGNAME, status_file);
		return;
	}
	
	sameLine = 0;
	while (fgets(line, sizeof(line), f) != NULL)
	{
		if (debugmode)
			fprintf(stdout, "%s: [get_printer_list] line = %s\n", PROGNAME, line);

		if (line[0] == '#')
			continue;
	
		if (sameLine)
			strcat(printerLine, line);
		else
			strcpy(printerLine, line);
		
		if (printerLine[strlen(printerLine) - 2] == '\\')
		{
			printerLine[strlen(printerLine) - 2] = '\0';
			sameLine = 1;
			continue;
		}
		else
		{
			sameLine = 0;
		}
		
		if (debugmode)
			fprintf(stdout, "%s: [get_printer_list] printerLine = %s\n", PROGNAME, printerLine);
		
		if ((charPtr = strchr(printerLine, ':')) == NULL)
			continue;
		else
			*charPtr = '\0';
			
		if (strlen(printerLine) > sizeof(name) - 1)
		{
			if (debugmode)
				fprintf(stdout, "%s: [get_printer_list] printer name = %s too long\n", PROGNAME, printerLine);	
			continue;
		}
		
		strcpy(name, printerLine);

		charPtr = strstr(charPtr + 1, "description=") + strlen("description=");

		if ((charPtr1 = strchr(charPtr, ':')) == NULL)
		{
			charPtr = NULL;
			comment[0] = '\0';
		}
		else
		{
			*charPtr1 = '\0';
			strcpy(comment, charPtr);
		}
		
		if (num_printers >= MAX_NUM_PRINTERS) 
		{
			if (debugmode)
				fprintf(stdout, "%s: [get_printer_list] Max number of printers (%d) exceeded.  Recompile with larger value\n", PROGNAME, MAX_NUM_PRINTERS);
			break;
		}
		
		if (debugmode)
		{
			fprintf(stdout, "%s: [get_printer_list] printers[%d] added\n", PROGNAME, num_printers);
			fprintf(stdout, "%s: [get_printer_list] queue = %s\n", PROGNAME, name);
			fprintf(stdout, "%s: [get_printer_list] comment = %s\n", PROGNAME, comment);
		}

		printers[num_printers].queuename = my_strdup(name);
		printers[num_printers].queuecomment = NULL;
		num_printers++;
	}

	fclose(f);
}

#endif
/*
 *	 Cache a list of available printers, BSD code searchs the /etc/printcap
 *	 file.	Creating a dummy /etc/printcap can also work.
 */
void
get_printers()
{
int i;
#ifndef FIND_QUEUES_SVR4
static char line[1024];
char *p, *p1, *end;
FILE *f;
int start;
#ifdef FIND_QUEUES_AIX
static char stanza[255];
#endif
#endif

	if (debugmode)
		fprintf(stdout, "%s: [get_printers] called\n", PROGNAME);

	num_printers = 0;
	if (DontDoPrinting)
		return;
#ifdef FIND_QUEUES_SVR4
	get_printer_list("/usr/spool/lp/system/pstatus", 0);
	get_printer_list("/usr/spool/lp/system/cstatus", 1);
#else
#ifdef FIND_QUEUES_SOLARIS26

	get_printer_list("/etc/printers.conf");

#else
#ifdef FIND_QUEUES_AIX
	if ((f=fopen("/etc/qconfig","r")) == NULL) {
		fprintf(stdout, "%s: File /etc/qconfig was not found.\n", PROGNAME);
		fprintf(stdout, "        No printers are available for use.\n");
		fflush(stdout);
		return;
	}

	stanza[0] = '\0';
	while (fgets(line, sizeof(line)-1, f) != NULL) {
		line[sizeof(line)-1] = '\0';
		for (p=line; isspace(*p); )
			p++;
		if (*p == '*')
			continue;
		for (p1=&line[strlen(line)-1]; isspace(*p1); --p1)
			*p1 = '\0';
		if (strchr(p, ':') != NULL) {
			strcpy(stanza, p);
			if (stanza[strlen(stanza)-1] == ':')
				stanza[strlen(stanza)-1] = '\0';
			continue;
		}
		if (my_strstr(p, "device") != NULL) {
			if (stanza[0] != '\0') {
				if (num_printers >= MAX_NUM_PRINTERS) {
					fprintf(stdout, "%s: [get_printers] Max number of printers (%d) exceeded.\nRecompile with larger value\n", PROGNAME, MAX_NUM_PRINTERS);
					fflush(stdout);
					goto outoutout;
				}
				if (strlen(stanza) <= 11) {
					printers[num_printers].queuename = my_strdup(stanza);
					printers[num_printers].queuecomment = NULL;
					printers[num_printers].queuestatus = WNPRQ_ACTIVE;
					printers[num_printers].numphysicalprinters = 0;
					printers[num_printers].lasttime = 0L;
					printers[num_printers].head = NULL;
					num_printers++;
				}
				else
					fprintf(stdout, "%s: [get_printer_list] Printer name \"%s\" exceeds 11 characters; not available for linking\n", PROGNAME, stanza);
				stanza[0] = '\0';
			}
		}
	}
outoutout:
	fclose(f);
#else
#ifdef FIND_QUEUES_PRINTCAP
	if ((f=fopen("/etc/printcap", "r")) == NULL) {
		fprintf(stdout, "%s: File /etc/printcap was not found.\n", PROGNAME);
		fprintf(stdout, "        No printers are available for use.\n");
		fflush(stdout);
		return;
	}

	while (fgets(line, sizeof(line)-1, f) != NULL) {
		start = num_printers;
multi_line:
		if ((p=strchr(line,'#')) != NULL)
			*p = '\0';
		if ((p=strchr(line,':')) != NULL)
			*p = '\0';
		if ((end=line+strlen(line)) == line)
			continue;
		if (*(end-1) == '\n')
			*(--end) = '\0';
		if (*(end-1) == '\\') {
			*(--end) = '\0';
			if ((sizeof(line)-(end-line)) <= 0) {
				fprintf(stdout, "%s: [get_printers] Multi-line line too long in /etc/printcap.\n", PROGNAME);
				fflush(stdout);
				break;
			}
			if (fgets(end, sizeof(line)-(end-line), f) != NULL)
				goto multi_line;
		}
		for (p=line; p<=end; ) {
			while ((*p == ' ') || (*p == '\t'))
				p++;
			if (!*p)
				break;
			p1 = p;
			while (strchr(" \t|", *p1) == NULL)
				p1++;
			if ((*p1 == ' ') || (*p1 == '\t')) {
				while (strchr("|", *p1) == NULL)
					p1++;
				*p1 = '\0';
				while ((p[strlen(p)-1] == ' ') || (p[strlen(p)-1] == '\t'))
					p[strlen(p)-1] = '\0';
				if ((strchr(p, ' ') != NULL) || (strchr(p, '\t') != NULL)) {
					for (i=start; i<num_printers; i++)
						printers[i].queuecomment = my_strdup(p);
					start = num_printers;
					p = p1 + 1;
					continue;
				}
			}
			else
				*p1 = '\0';
			if (*p) {
				if (num_printers >= MAX_NUM_PRINTERS) {
					fprintf(stdout, "%s: [get_printers] Max number of printers (%d) exceeded.\nRecompile with larger value\n", PROGNAME, MAX_NUM_PRINTERS);
					fflush(stdout);
					goto outoutout;
				}
				if (strlen(p) <= 11) {
					printers[num_printers].queuename = my_strdup(p);
					printers[num_printers].queuecomment = NULL;
					printers[num_printers].queuestatus = WNPRQ_ACTIVE;
					printers[num_printers].numphysicalprinters = 0;
					printers[num_printers].lasttime = 0L;
					printers[num_printers].head = NULL;
					num_printers++;
				}
				else
					fprintf(stdout, "%s: [get_printer_list] Printer name \"%s\" exceeds 11 characters; not available for linking\n", PROGNAME, p);
			}
			p = p1 + 1;
		}
	}
outoutout:
	fclose(f);
#else
 error FIND_QUEUES_Something must be specified
#endif
#endif
#endif
#endif
	if (debugmode) {
		  fprintf(stdout, "%s: [get_printers] Following devices are available:\n", PROGNAME);
		  for (i=0; i<num_printers; i++)
			  fprintf(stdout, "%s:                 %s\n", PROGNAME, printers[i].queuename);
	}
}

static struct var_file {
	int status;
	int dir_size;
	char *dir_handle;
} f;

/*
 *	 DOS 3.1 Sharing and Unsharing code. Uses the same format for rpc.lockd
 *	 as HCLNFSD, so some parameters are ignored. The cookie under rpc.lockd
 *	 must be 4 bytes long under HCLNFSD.
 */
xdr_share(xdrsp, l)
	XDR *xdrsp;
	struct locking *l;
{
unsigned int i;
char g[32], *p;

	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (i != 4)
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->cookie = i;

	p = l->name;
	if (!xdr_string(xdrsp, &p, sizeof(l->name)))
		return(0);
	p = l->fh;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(l->fh)))
		return(0);
	p = g;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(g)))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->mode = i;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->access = i;
	if (debugmode)
		fprintf(stdout, "%s: [xdr_share] returns Success\n", PROGNAME);
	if (debugmode > 1) {
	unsigned long t1=l->offset, t2=l->length, t3=l->stat, t4=l->sequence;
		fprintf(stdout, "%s: [xdr_share] called for <%s>, ofs = %lu, len = %lu\n", PROGNAME, l->name, t1, t2);
		t1 = l->mode;
		t2 = l->access;
		fprintf(stdout, "%s:             mode = %lu, access = %lu, stat = %lu, sequence = %lu\n", PROGNAME, t1, t2, t3, t4);
	}
	return(1);
}


/*
 *	 Get Printers response
 */
int
xdr_printers(xdrsp, l)
	XDR *xdrsp;
	char *l;
{
int i;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_printers] called\n", PROGNAME);
	i = num_printers;
	if (!xdr_int(xdrsp, &i))
		return(0);
	for (i=0; i<num_printers; i++) {
		p = printers[i].queuename;
		if (strlen(p) > 24) {
			*(p+24) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [xdr_printers] queuename truncated to <%s>\n", PROGNAME, p);
		}
		if (!xdr_string(xdrsp, &p, 24))
			return(0);
		p = printers[i].queuecomment;
		if (p == NULL)
			p = "";
		if (strlen(p) > 255) {
			*(p+255) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [xdr_printers] queuecomment truncated to <%s>\n", PROGNAME, p);
		}
		if (!xdr_string(xdrsp, &p, 255))
			return(0);
	}

	if (debugmode)
		fprintf(stdout, "%s: [xdr_printers] returns Success\n", PROGNAME);
	return(1);
}

/*
 *	 DOS 3.1 Share response.
 */
int
xdr_shareres(xdrsp,l)
	XDR *xdrsp;
	struct locking *l;
{
unsigned int i;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_shareres] called by <%s>\n", PROGNAME, l->name);
	i = 4;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = l->cookie;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = l->stat;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->sequence++;
	i = l->sequence;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);

	if (debugmode)
		fprintf(stdout, "%s: [xdr_shareres] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 DOS 3.1 locking request. Note that several parameters are ignored.
 */
int
xdr_lock(xdrsp, l)
	XDR *xdrsp;
	struct locking *l;
{
unsigned int i;
union {
   char g[32];
   _uint32 lockowner;
} g;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_lock] called by <%s>\n", PROGNAME, l->name);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->cookie = i;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->exclusive = i;
	p = l->name;
	if (!xdr_string(xdrsp, &p, sizeof(l->name)))
		return(0);
	if (strlen(p) > 17) {
		*(p+17) = '\0';
		if (debugmode)
			fprintf(stdout, "%s: [xdr_lock] name truncated to <%s>\n", PROGNAME, p);
	}
	p = l->fh;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(l->fh)))
		return(0);
	p = (char *)&g;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(g)))
		return(0);
	l->lockowner = g.lockowner;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->offset = i;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->length = i;
	if (debugmode > 1) {
	unsigned long t1=l->offset, t2=l->length;
		fprintf(stdout, "%s: [xdr_lock] returns Success\n", PROGNAME);
		fprintf(stdout, "%s:            ofs = %lu, len = %lu\n", PROGNAME, t1, t2);
	}
	return(1);
}


/*
 *	 DOS 3.1 unlock request. Note that several parameters are ignored.
 */
xdr_unlock(xdrsp, l)
	XDR *xdrsp;
	struct locking *l;
{
unsigned int i;
char g[32], *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_unlock] called by <%s>\n", PROGNAME, l->name);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->cookie = i;

	p = l->name;
	if (!xdr_string(xdrsp, &p, sizeof(l->name)))
		return(0);
	if (strlen(p) > 17) {
		*(p+17) = '\0';
		if (debugmode)
			fprintf(stdout, "%s: [xdr_unlock] name truncated to <%s>\n", PROGNAME, p);
	}
	p = l->fh;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(l->fh)))
		return(0);
	p = g;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(g)))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->offset = i;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	l->length = i;
	if (debugmode > 1) {
	unsigned long t1=l->offset, t2=l->length;
		fprintf(stdout, "%s: [xdr_unlock] returns Success\n", PROGNAME);
		fprintf(stdout, "%s:              ofs = %lu, len = %lu\n", PROGNAME, t1, t2);
	}
	return(1);
}

/*
 *	 Send Locking response.
 */
int
xdr_res(xdrsp, l)
	XDR *xdrsp;
	struct locking *l;
{
unsigned int i;
char *p;

	if (debugmode > 1)
		fprintf(stdout, "%s: [xdr_res] called\n", PROGNAME);
	i = 4;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = l->cookie;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = l->stat;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_res] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 Remove all DOS 3.1 locks and shares, note that the PC's name is
 *	 a maximum of 16 chaarcters.
 */
int
xdr_get_lock_name(xdrsp, l)
	XDR *xdrsp;
	char *l;
{
	if (debugmode)
		fprintf(stdout, "%s: [xdr_get_lock_name] called\n", PROGNAME);
	return(xdr_string(xdrsp, &l, 17));
}

/*
 *	 Send a redirect to a new IP during an authorization request.
 */
int
xdr_new_ip(xdrsp, ip)
	XDR *xdrsp;
	_uint32 *ip;
{
unsigned int i;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_new_ip] called\n", PROGNAME);
	i = 3;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = *ip;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_new_ip] returns Success\n", PROGNAME);
	return(1);
}

/*
 *	 Get GROUP name.
 */
int
xdr_g_string(xdrsp, st)
	XDR *xdrsp;
	char *st;
{
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_g_string] called\n", PROGNAME);
	p = st;
	if (!xdr_string(xdrsp, &p, 32))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_g_string] returns Success\n", PROGNAME);
	return(1);
}

/*
 *	 Send Status and GID
 */
int
xdr_send_gid(xdrsp, st)
	XDR *xdrsp;
	char *st;
{
int *i;
unsigned int j;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_gid] called with st = <%s>\n", PROGNAME, st);
	i = get_gr_gid(st);
	if (i == NULL) {
		j = 2;
		if (!xdr_int(xdrsp, (int *)&j))
			return(0);
	}
	else {
		j = 0;
		if (!xdr_int(xdrsp, (int *)&j))
			return(0);
		if (!xdr_int(xdrsp, i))
			return(0);
	}
	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_gid] returns Success\n", PROGNAME);
	return(1);
}

/*
 *	 Get list of UIDs to translate to text.
 */
int
xdr_nuids(xdrsp, luids)
	XDR *xdrsp;
	int luids[NUSRS+1];
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_nuids] called\n", PROGNAME);
	if (!xdr_int(xdrsp, &luids[0]))
		return(0);
	if ((luids[0] < 0) || (luids[0] > 16))
		luids[0] = 0;
	for (i=1; i<=luids[0]; i++)
	   if (!xdr_int(xdrsp, &luids[i]))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_nuids] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 Get list of GIDs to translate to text.
 */
int
xdr_ngids(xdrsp, lgids)
	XDR *xdrsp;
	int lgids[NGRPS+1];
{
int i;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_ngids] called\n", PROGNAME);

	if (!xdr_int(xdrsp, &lgids[0]))
		return(0);
	if ((lgids[0] < 0) || (lgids[0] > 16))
		lgids[0] = 0;
	for (i=1; i<=lgids[0]; i++)
		if (!xdr_int(xdrsp, &lgids[i]))
			return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_ngids] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 Send Usernames translated from UIDS.
 */
xdr_send_usernames(xdrsp, luids)
	XDR *xdrsp;
	int luids[NUSRS+1];
{
int i, j;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_usernames] called\n", PROGNAME);
	for (i=1; i<=luids[0]; i++) {
		p = get_ui_name(luids[i]);
		if (debugmode)
			fprintf(stdout, "%s: [xdr_send_usernames] p = <%s>\n", PROGNAME, p);
		if (p == NULL) {
			j = 0;
			if (!xdr_int(xdrsp, &j))
				return(0);
		}
		else {
			if (strlen(p) > 32) {
				*(p+32) = '\0';
				if (debugmode)
					fprintf(stdout, "%s: [xdr_send_usernames] name truncated to <%s>\n", PROGNAME, p);
			}
			if (!xdr_string(xdrsp, &p, 32))
				return(0);
		}
	}
	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_usernames] returns Success\n", PROGNAME);
	return(1);
}

/*
 *	 Send Group names translated from GIDS.
 */
int
xdr_send_names(xdrsp, lgids)
	XDR *xdrsp;
	int lgids[NGRPS+1];
{
int i, j;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_names] called\n", PROGNAME);
	for (i=1; i<=lgids[0]; i++) {
		p = get_gr_name(lgids[i]);
		if (debugmode)
			fprintf(stdout, "%s: [xdr_send_names] p = <%s>\n", PROGNAME, p);
		if (p == NULL) {
			j = 0;
			if (!xdr_int(xdrsp, &j))
				return(0);
		}
		else {
			if (strlen(p) > 32) {
				*(p+32) = '\0';
				if (debugmode)
					fprintf(stdout, "%s: [xdr_send_names] name truncated to <%s>\n", PROGNAME, p);
			}
			if (!xdr_string(xdrsp, &p, 32))
				return(0);
		}
	}
	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_names] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 Return fhandle of spool directory. Note a loop of XDR_INTS is
 *	 used to allow this to compile on all systems.
 *
 */
xdr_file(xdrsp, file_var)
	XDR *xdrsp;
	struct var_file *file_var;
{
unsigned int i,j;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_file] called\n", PROGNAME);
	if (!xdr_int(xdrsp, &file_var->status))
		return(0);
	if (file_var->status != 0) {
		if (debugmode > 1)
			fprintf(stdout, "%s: [xdr_file] status = %u, returns Success\n", PROGNAME, file_var->status);
		return(1);
	}
	for (i=0; i<32; i+=4) {
		j = *(_uint32 *)((&file_var->dir_handle[0]) + i);
		if (!xdr_int(xdrsp, (int *)&j))
			return(0);
	}
	if (debugmode)
		fprintf(stdout, "%s: [xdr_file] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	Spool a file to the printer:
 *
 *		Parameters:
 *			printer_name
 *			file_name    (fixed 8 bytes) (Internet Address in Hex).
 *			extension     xdr_int (filename is form C02A0405.xxx)
 */
int
xdr_print(xdrsp)
	XDR *xdrsp;
{
struct stat buf, buf1;
struct passwd *pw1;
char *p;
unsigned int i;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_print] called\n", PROGNAME);
	if (DontDoPrinting) {
		do_print = -1;
		return(1);
	}
	i = sizeof(printer)-1;
	p = printer;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(printer)-1)) {
		if (debugmode > 1)
			fprintf(stdout, "%s: [xdr_print] Error getting printer name\n", PROGNAME);
		return(0);
	}
	printer[i] = ' ';
	*strchr(printer, ' ') = '\0';
	i = 8;
	p = file_name;
	if (!xdr_bytes(xdrsp, &p, &i, 8)) {
		if (debugmode > 1)
			fprintf(stdout, "%s: [xdr_print] Error getting filename; printer = %s\n", PROGNAME, printer);
		return(0);
	}
	file_name[8] = '\0';
	if (!xdr_int(xdrsp, &extension)) {
		if (debugmode > 1)
			fprintf(stdout, "%s: [xdr_print] Error getting extension; printer = %s, file = %s\n", PROGNAME, printer, file_name);
		return(0);
	}
	strlwr(file_name);
	sprintf(file, "%s/%s.%03d", spool_dir, file_name, extension);
avoid_slowprinters:
#ifdef FOURTEEN_CHAR
	sprintf(file1, "%s/%s.%03dq", spool_dir, file_name, extension);
#else
	sprintf(file1,"%s/%s.%03dqueued", spool_dir, file_name, extension);
#endif
	if (debugmode > 1)
		fprintf(stdout, "%s: [xdr_print] f = %s, f1 = %s\n", PROGNAME, file, file1);
	jobname = NULL;
	do_print = -1;
	if (stat(file, &buf) == 0) {
		if (stat(file1, &buf1) == 0) {
			extension += 100;
			if (debugmode)
				fprintf(stdout, "%s: [xdr_print] printfile name collision\n", PROGNAME);
			goto avoid_slowprinters;
		}
		do_print = rename(file, file1);
		if (debugmode && do_print)
			fprintf(stdout, "%s: [xdr_print] rename failed, error = %d\n", PROGNAME, errno);
		endpwent();
		pw1 = getpwuid((int) buf.st_uid);
		if (pw1 != NULL) {
			if (debugmode)
				fprintf(stdout, "%s: [xdr_print] jobname = %s\n", PROGNAME, pw1->pw_name);
			jobname = pw1->pw_name;
		}
	}
	else {
		if (debugmode) {
			if (errno == ENOENT)
				fprintf(stdout, "%s: [xdr_print] duplicate close request\n", PROGNAME);
			else
				fprintf(stdout, "%s: [xdr_print] stat failed, errno = %d\n", PROGNAME, errno);
		}
	}
	if (debugmode)
		fprintf(stdout, "%s: [xdr_print] returns Success\n", PROGNAME);
	return(1);
}

char *
find_host(r_ip)
	_uint32 r_ip;
{
char *p;
_uint32 incomming_ip;
struct hostent *h;

	if (r_ip == my_ip)
		p = hostname;
	else {
		incomming_ip = (_uint32)ntohl(r_ip);
		h = gethostbyaddr((char *)&incomming_ip, 4, AF_INET);
		if (h == NULL) {
			sprintf(remote_host,"%u.%u.%u.%u",
				       (r_ip >> 24) & 0xff,
				       (r_ip >> 16) & 0xff,
				       (r_ip >>  8) & 0xff,
				       (r_ip	  ) & 0xff);
			p = remote_host;
		}
		else
			p = h->h_name;
	}
	return(p);
}

struct auth_struct {
	_uint32 ip;
	_uint32 type;
	char device[65];
	char login_text[64];
	char *username;
	char *password;
} auth_request;

/*
 *	 Request Authorization:
 *		 Parameters:
 *			 IP_Address of server host (XDR_INT)
 *			 Type of request (XDR_INT) (4 - disk, 3 - printer)
 *			 Device to link  (XDR_BYTES max 12) (printer name)
 *			 Username/Password (XDR_BYTES max 64)
 */
int
xdr_auth_request(xdrsp, auth)
	XDR *xdrsp;
	struct auth_struct *auth;
{
unsigned int i;
int j, x, y;
char *p;

	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	auth->ip = i;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	auth->type = i;
	i = sizeof(auth->device)-1;
	p = auth->device;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(auth->device)-1))
		return(0);
	auth->device[i] = ' ';
	*strchr(auth->device, ' ') = '\0';
	i = sizeof(auth->login_text);
	p = auth->login_text;
	if (!xdr_bytes(xdrsp, &p, &i, sizeof(auth->login_text)))
		return(0);
	if (i < 6)
		auth->type = -1;
	p = auth->login_text;
	for (x= -1,j=0; j<i; j++) {
		y = *p;
		x ^= *p;
		*p++ = x;
		x = y;
	}
	auth->username = auth->login_text + 2;
	auth->password = auth->username + strlen(auth->username) + 1;
	if (debugmode) {
	unsigned long t1=auth->type;
		fprintf(stdout, "%s: [xdr_auth_request] called\n", PROGNAME);
		fprintf(stdout, "%s:   Link to %d.%d.%d.%d (%s)\n", PROGNAME,
			   (auth->ip >> 24) & 0xff, (auth->ip >> 16) & 0xff,
			   (auth->ip >>  8) & 0xff, (auth->ip)	     & 0xff,
			   find_host(auth->ip));
		fprintf(stdout, "%s:   Request type     : %lu (%s)\n", PROGNAME, t1, (auth->type==3)? "print request": "disk request");
		fprintf(stdout, "%s:   Requested Device : %s\n", PROGNAME, auth->device);
		fprintf(stdout, "%s:   Username         : %s\n", PROGNAME, auth->username);
/*
 *		fprintf(stdout, "%s:   Password         : %s\n", PROGNAME, auth->password);
 */
		fprintf(stdout, "%s: [xdr_auth_request] returns Success\n", PROGNAME);
	}
	return(1);
}

struct name_struct {
	char usr_login[64];
	char *usr_name;
} name_request;

int
xdr_name_request(xdrsp, name)
	XDR *xdrsp;
	struct name_struct *name;
{
char *p;

	p = name->usr_login;
	name->usr_name = name->usr_login;
	if (!xdr_string(xdrsp, &p, sizeof(name->usr_login)))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_name_request] called Successfully\n", PROGNAME);
	return(1);
}

/*
 *	 Send successful authorization reply.
 *		 Results:
 *		   UID (XDR_INT)
 *		   GID (XDR_INT)
 *		   GID count (XDR_INT)
 *		   GIDs (XDR_INT)	 (GID count entries)
 */
int
xdr_pw(xdrsp, pwd)
	XDR *xdrsp;
	struct passwd *pwd;
{
unsigned int i;
struct mm_gidss *gg;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_pw] called\n", PROGNAME);
	gg = get_gids(pwd->pw_uid, pwd->pw_gid, pwd->pw_name);
	if (debugmode > 1) {
		fprintf(stdout, "%s: [xdr_pw] uid = %u, gid = %u, gidcount = %u\n", PROGNAME, gg->mm_uid, gg->mm_gid, gg->mm_gid_count);
		for (i=0; i<gg->mm_gid_count; i++)
			fprintf(stdout, "%s: [xdr_pw] gids[%d] = %u\n", PROGNAME, i, gg->mm_gids[i]);
	}
	i = 0;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = pwd->pw_uid;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	i = pwd->pw_gid;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (!xdr_int(xdrsp, &gg->mm_gid_count))
		return(0);
	for (i=0; i<gg->mm_gid_count; i++)
		if (!xdr_int(xdrsp, &gg->mm_gids[i]))
			return(0);

	if (debugmode > 1)
		fprintf(stdout, "%s: [xdr_pw] returns Success\n", PROGNAME);
	return(1);
}

int
xdr_get_printq(xdrsp,pq)
	XDR *xdrsp;
	struct printqs *pq;
{
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_get_printq] called\n", PROGNAME);
	p = pq->queuename;
	if (!xdr_string(xdrsp, &p, sizeof(pq->queuename)))
		return(0);
	p = pq->username;
	if (!xdr_string(xdrsp, &p, sizeof(pq->username)))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_get_printq] returns Success\n", PROGNAME);
	return(1);
}

int
xdr_cancel_job(xdrsp, pq)
	XDR *xdrsp;
	struct printqs *pq;
{
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_cancel_job] called\n", PROGNAME);
	p = pq->queuename;
	if (!xdr_string(xdrsp, &p, sizeof(pq->queuename)))
		return(0);
	p = pq->username;
	if (!xdr_string(xdrsp, &p, sizeof(pq->username)))
		return(0);
	if (!xdr_int(xdrsp,&pq->printer_number))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_cancel_job] returns Success\n", PROGNAME);
	return(1);
}

int
xdr_send_printq(xdrsp, pq)
	XDR *xdrsp;
	struct printqs *pq;
{
unsigned int i;
int pn;
struct queueentry *qe;
char *pp;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_printq] called\n", PROGNAME);
	pn = pq->printer_number;
	i = (pn == -1);
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (i != 0)
		return(1);
	pp = printers[pn].queuecomment;
	if (pp == NULL)
		pp = "";
	if (strlen(pp) > 32) {
		*(pp+32) = '\0';
		if (debugmode)
			fprintf(stdout, "%s: [xdr_send_printq] queuecomment truncated to <%s>\n", PROGNAME, pp);
	}
	if (!xdr_string(xdrsp, &pp, 32))
		return(0);
	if (!xdr_int(xdrsp, &printers[pn].queuestatus))
		return(0);
	if (!xdr_int(xdrsp, &printers[pn].numphysicalprinters))
		return(0);
	qe = printers[pn].head;
	i = 1;
	while (qe) {
		if ((*pq->username != '\0') && (strcmp(pq->username, qe->username) != 0)) {
			qe = qe->next;
			continue;
		}
		if (!xdr_int(xdrsp, (int *)&i))
			return(0);
		if (!xdr_int(xdrsp, &qe->jobid))
			return(0);
		pp = qe->username;
		if (pp == NULL)
			pp = "";
		if (strlen(pp) > 32) {
			*(pp+32) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [xdr_send_printq] username truncated to <%s>\n", PROGNAME, pp);
		}
		if (!xdr_string(xdrsp, &pp, 32))
			return(0);
		pp = qe->params;
		if (pp == NULL)
			pp = "";
		if (strlen(pp) > 255) {
			*(pp+255) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [xdr_send_printq] params truncated to <%s>\n", PROGNAME, pp);
		}
		if (!xdr_string(xdrsp, &pp, 255))
			return(0);
		if (!xdr_int(xdrsp, &qe->queueposition))
			return(0);
		if (!xdr_int(xdrsp, &qe->jobstatus))
			return(0);
		if (!xdr_long(xdrsp, (long *)&qe->timesubmitted))
			return(0);
		if (!xdr_long(xdrsp, (long *)&qe->size))
			return(0);
		if (!xdr_int(xdrsp, &qe->copies))
			return(0);
		pp = qe->jobcomment;
		if (pp == NULL)
			pp = "";
		if (strlen(pp) > 255) {
			*(pp+255) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [xdr_send_printq] jobcomment truncated to <%s>\n", PROGNAME, pp);
		}
		if (!xdr_string(xdrsp, &pp, 255))
			return(0);
		qe = qe->next;
	}
	i = 0;
	if (!xdr_int(xdrsp, (int *)&i))
		return(0);
	if (debugmode)
		fprintf(stdout, "%s: [xdr_send_printq] returns Success\n", PROGNAME);
	return(1);
}

#ifdef SCO
#define SecureWare 1
#include <sys/security.h>
#include <sys/audit.h>
#include <prot.h>
#endif

static struct user_cache_struct {
	char username[32];
	int uid, gid;
	char password[32];
	unsigned long successes;
} user_cache[NUM_USER_CACHE] = { 0 };

static int cache_cycle=0;

static void
add_cache(pw, p)
	struct passwd *pw;
	char *p;
{
#ifndef NO_USERNAME_CACHE
int i, pos;
unsigned long smin= (unsigned long)~0L;

	pos = 0;
	cache_cycle = (cache_cycle+1) % NUM_USER_CACHE;
	for (i=0; i<NUM_USER_CACHE; i++)
		if ((user_cache[i].successes < smin) ||
		    ((user_cache[i].successes == smin) && (i <= cache_cycle))) {
			pos = i;
			smin = user_cache[i].successes;
		}
	if (debugmode)
		fprintf(stdout, "%s: [addcache] <%s> added to position %d replacing <%s> (%ld hits)\n", PROGNAME,
			pw->pw_name, pos, (user_cache[pos].username[0]=='\0')? "empty": user_cache[pos].username,
			user_cache[pos].successes);
	strcpy(user_cache[pos].username, pw->pw_name);
	strcpy(user_cache[pos].password, p);
	user_cache[pos].uid = pw->pw_uid;
	user_cache[pos].gid = pw->pw_gid;
	user_cache[pos].successes = 1;
#endif
}

struct passwd *
getcache(request)
	struct auth_struct *request;
{
#ifndef NO_USERNAME_CACHE
static struct passwd pas;
int i;

	for (i=0; i<NUM_USER_CACHE; i++) {
		if (strcmp(request->username, user_cache[i].username) == 0) {
			if (strcmp(crypt(request->password, user_cache[i].password), user_cache[i].password) == 0) {
				pas.pw_name = user_cache[i].username;
				pas.pw_uid = user_cache[i].uid;
				pas.pw_gid = user_cache[i].gid;
				user_cache[i].successes++;
				if (debugmode)
					fprintf(stdout, "%s: [getcache] <%s> found in cache (%ld hits)\n", PROGNAME, request->username, user_cache[i].successes);
				return(&pas);
			}
			else {
				user_cache[i].username[0] = '\0';
				user_cache[i].successes = 0;
				if (debugmode)
					fprintf(stdout, "%s: [getcache] <%s> requested password doesn't match password in cache\n", PROGNAME, request->username);
				return(NULL);
			}
		}
	}
#endif
	return(NULL);
}

char *
reqname(req)
	int req;
{
static char buff[100];

	switch(req) {
	case NULLPROC:
		strcpy(buff, "Null Procedure");
		break;
	case SPOOL_INQUIRE:
		strcpy(buff, "Return fhandle of spool directory");
		break;
	case SPOOL_FILE:
		strcpy(buff, "Spool a file");
		break;
	case AUTHORIZE:
		strcpy(buff, "Authorize and return UID and GIDs");
		break;
	case GRP_NAME_TO_NUMB:
		strcpy(buff, "Convert group name to number");
		break;
	case GRP_TO_NUMBER:
		strcpy(buff, "Convert group number(s) to name(s)");
		break;
	case RETURN_HOST:
		strcpy(buff, "Convert IP to hostname");
		break;
	case UID_TO_NAME:
		strcpy(buff, "Convert UID(s) to name(s)");
		break;
	case NAME_TO_UID:
		strcpy(buff, "Convert name(s) to UID(s)");
		break;
	case SHARE:
		strcpy(buff, "DOS Share function");
		break;
	case UNSHARE:
		strcpy(buff, "DOS UnShare function");
		break;
	case ZAP_LOCKS:
		strcpy(buff, "Remove all locks on file by process");
		break;
	case LOCK:
		strcpy(buff, "DOS Lock request");
		break;
	case REMOVE:
		strcpy(buff, "Remove all locks/shares for PC");
		break;
	case UNLOCK:
		strcpy(buff, "DOS Unlock request");
		break;
	case GET_PRINTERS:
		strcpy(buff, "Get a list of printers");
		break;
	case GET_PRINTQ:
		strcpy(buff, "Get print queue entries and status");
		break;
	case CANCEL_PRJOB:
		strcpy(buff, "Cancel a print job");
		break;
	default:
		sprintf(buff, "Unknown request - %d", req);
		break;
	}
	return(buff);
}

/*
 *	 The main RPC routine.
 */
void
create_spool(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
struct passwd *pw;
int i, j;
_uint32 r_ip, i_ip;
char *p;
#ifdef SHADW
struct spwd *sp;
#endif
#ifdef SCO
struct pr_passwd *prpw;
#endif

	if (debugmode) {
		fprintf(stdout, "%s: [create_spool] called with %s\n", PROGNAME, reqname(rqstp->rq_proc));
		fflush(stdout);
	}
	switch(rqstp->rq_proc) {
	case NULLPROC:
		if (!svc_sendreply(transp, (xdrproc_t)xdr_void, (Generic_t)NULL))
			if (debugmode)
				fprintf(stdout, "%s: [nullproc] Can't reply to RPC call\n", PROGNAME);
		return;
	case SPOOL_INQUIRE:
		if (!svc_sendreply(transp, (xdrproc_t)xdr_file, (Generic_t)&f))
			if (debugmode)
				fprintf(stdout, "%s: [spool_inquire] Can't reply to RPC call\n", PROGNAME);
		return;
	case SPOOL_FILE:
		if (!svc_getargs(transp, (xdrproc_t)xdr_print, (Generic_t)NULL)) {
			if (debugmode)
				fprintf(stdout, "%s: [spool_file] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		i = 0;
		if (do_print == 0) {
			for (i=0; i<num_printers; i++)
				if (strcmp(printers[i].queuename, printer) == 0) {
					printers[i].lasttime = 0L;
					break;
				}
			if (debugmode)
				fflush(stdout);
			if ((i=fork()) == 0) {
				(void) signal(SIGHUP,SIG_IGN);
				(void) signal(SIGINT,SIG_IGN);
				(void) signal(SIGQUIT,SIG_IGN);
#ifdef SIGABRT
				(void) signal(SIGABRT,SIG_IGN);
#endif
				(void) signal(SIGTERM,SIG_IGN);
				print_it(file1,printer,jobname);
				exit(0);
			}
		}
		if (i != -1) {
			if (!svc_sendreply(transp, (xdrproc_t)xdr_void, (Generic_t)NULL))
				if (debugmode)
					fprintf(stdout, "%s: [spool_file] Can't reply to RPC call\n", PROGNAME);
		}
		else
			if (debugmode)
				fprintf(stdout, "%s: [spool_file] Can't fork print process\n", PROGNAME);
		return;
	case AUTHORIZE:
		if (!svc_getargs(transp, (xdrproc_t)xdr_auth_request, (Generic_t)&auth_request)) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		if (auth_request.type == -1) {
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [authorize] (1) Can't reply to RPC call\n", PROGNAME);
			return;
		}
		if (A_flag == 0) {
			if (s_flag == 0) {
				if ((_uint32)ntohl(auth_request.ip) != my_ip) {
					i = 1;
					if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
						if (debugmode)
							fprintf(stdout, "%s: [authorize] (2) Can't reply to RPC call\n", PROGNAME);
					return;
				}
			}
			else {
				for (i=0; i<ips_count; i++) {
					if (((*(ips+i))->mask & (_uint32)ntohl(auth_request.ip)) == (*(ips+i))->ip) {
						switch((*(ips+i))->type) {
						case '+' :
							i = ips_count+1;
							break;
						case '-' :
							i = 1;
							if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
								if (debugmode)
									fprintf(stdout, "%s: [authorize] (3) Can't reply to RPC call\n", PROGNAME);
							return;
						case '=' :
							i_ip = (*(ips+i))->redirect_ip;
							if (!svc_sendreply(transp, (xdrproc_t)xdr_new_ip, (Generic_t)&i_ip))
								if (debugmode)
									fprintf(stdout, "%s: [authorize] (4) Can't reply to RPC call\n", PROGNAME);
							return;
						}
					}
				}
				if (i == ips_count) {
					i = 1;
					if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
						if (debugmode)
							fprintf(stdout, "%s: [authorize] (5) Can't reply to RPC call\n", PROGNAME);
					return;
				}
			}
		}
#ifdef SCO
		if ((prpw=getprpwnam(auth_request.username)) == NULL) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] getprpwnam failed\n", PROGNAME);
auth_error:
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [authorize] (6) Can't reply to RPC call\n", PROGNAME);
			return;
		}
		if (prpw->uflg.fg_lock && (prpw->ufld.fd_lock != 0))
			goto auth_error;
		i = prpw->ufld.fd_max_tries;
		if (i == 0)
			i = 5;
		if (prpw->ufld.fd_nlogins >= (short)i)
			goto auth_error;
		prpw->ufld.fd_encrypt[13] = '\0';
		if (strcmp(crypt(auth_request.password, prpw->ufld.fd_encrypt), prpw->ufld.fd_encrypt) != 0) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] crypt(pwd)=%s doesn't match %s\n", PROGNAME, crypt(auth_request.password, prpw->ufld.fd_encrypt), prpw->ufld.fd_encrypt);
			goto auth_error;
		}
		endpwent();
		if ((pw=getpwnam(auth_request.username)) == NULL) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] getpwnam failed\n", PROGNAME);
			goto auth_error;
		}
#else
		if ((pw=getcache(&auth_request)) != NULL)
			goto got_pw;
		endpwent();
		if ((pw=getpwnam(auth_request.username)) == NULL) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] getpwnam failed\n", PROGNAME);
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [authorize] (6) Can't reply to RPC call\n", PROGNAME);
			return;
		}
#ifdef SHADW
		if ((sp=getspnam(auth_request.username)) == NULL) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] getspnam request failed\n", PROGNAME);
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [authorize] (7) Can't reply to RPC call\n", PROGNAME);
			return;
		}
		p = sp->sp_pwdp;
#else
		p = pw->pw_passwd;
#endif
		if (strcmp(crypt(auth_request.password, p), p) != 0) {
			if (debugmode)
				fprintf(stdout, "%s: [authorize] crypt(pwd)=%s doesn't match %s\n", PROGNAME, crypt(auth_request.password,p), p);
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [authorize] (8) Can't reply to RPC call\n", PROGNAME);
			return;
		}
		add_cache(pw, p);
#endif
got_pw:
		if (auth_request.type == 3) {
			for (i=0; i<num_printers; i++)
				if (strcmp(printers[i].queuename, auth_request.device) == 0)
					break;
			if (i == num_printers) {
				if (debugmode)
					fprintf(stdout, "%s: [authorize] Printer %s not found\n", PROGNAME, auth_request.device);
				i = 2;
				if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
					if (debugmode)
						fprintf(stdout, "%s: [authorize] (9) Can't reply to RPC call\n", PROGNAME);
				return;
			}
#ifdef CHECK_PRINTER_ACCESS
		       if (!check_access(i, auth_request.username)) {
				if (debugmode)
					fprintf(stdout, "%s: [authorize] Not authorized for %s\n", PROGNAME, auth_request.device);
				i = 1;
				if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
					if (debugmode)
						fprintf(stdout, "%s: [authorize] (10) Can't reply to RPC call\n", PROGNAME);
				return;
			}
#endif
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_pw, (Generic_t)pw))
			if (debugmode)
				fprintf(stdout, "%s: [authorize] (11) Can't reply to RPC call\n", PROGNAME);
		return;
	case GRP_NAME_TO_NUMB:
		if (!svc_getargs(transp, (xdrproc_t)xdr_g_string, (Generic_t)g_string)) {
			if (debugmode)
				fprintf(stdout, "%s: [grp_name_to_numb] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_send_gid, (Generic_t)g_string))
			if (debugmode)
				fprintf(stdout, "%s: [grp_name_to_numb] Can't reply to RPC call\n", PROGNAME);
		return;
	case GRP_TO_NUMBER:
		if (!svc_getargs(transp, (xdrproc_t)xdr_ngids, (Generic_t)lgids)) {
			if (debugmode)
				fprintf(stdout, "%s: [grp_to_number] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_send_names, (Generic_t)lgids))
			if (debugmode)
				fprintf(stdout, "%s: [grp_to_number] Can't reply to RPC call\n", PROGNAME);
		return;
	case RETURN_HOST:
		if (!svc_getargs(transp, (xdrproc_t)xdr_int, (Generic_t)&i)) {
			if (debugmode)
				fprintf(stdout, "%s: [return_host] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		r_ip = i;
		p = find_host(r_ip);
		if (debugmode)
			fprintf(stdout, "%s: [return_host] remote = %s\n", PROGNAME, p);

		if (strlen(p) > 32) {
			*(p+32) = '\0';
			if (debugmode)
				fprintf(stdout, "%s: [return_host] string truncated to %s\n", PROGNAME, p);
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_g_string, (Generic_t)p))
			if (debugmode)
				fprintf(stdout, "%s: [return_host] Can't reply to RPC call\n", PROGNAME);
		return;
	case UID_TO_NAME:
		if (!svc_getargs(transp, (xdrproc_t)xdr_nuids, (Generic_t)luids)) {
			if (debugmode)
				fprintf(stdout, "%s: [uid_to_name] svc_getargs failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_send_usernames, (Generic_t)luids))
			if (debugmode)
				fprintf(stdout, "%s: [uid_to_name] Can't reply to RPC call\n", PROGNAME);
		return;
	case NAME_TO_UID:
		if (!svc_getargs(transp, (xdrproc_t)xdr_name_request, (Generic_t)&name_request)) {
			if (debugmode)
				fprintf(stdout, "%s: [name_to_uid] svc_getargs failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		endpwent();
		if ((pw=getpwnam(name_request.usr_name)) == NULL) {
			i = 1;
			if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&i))
				if (debugmode)
					fprintf(stdout, "%s: [name_to_uid] (1) Can't reply to RPC call\n", PROGNAME);
			return;
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_pw, (Generic_t)pw))
			if (debugmode)
				fprintf(stdout, "%s: [name_to_uid] (2) Can't reply to RPC call\n", PROGNAME);
		return;
	case REMOVE:
		if (!svc_getargs(transp, (xdrproc_t)xdr_get_lock_name, (Generic_t)lock.name)) {
			if (debugmode)
				fprintf(stdout, "%s: [remove] svc_getargs failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		process_lock_request((int)rqstp->rq_proc, lock.name, lock.fh, lock.access, lock.mode, lock.offset, lock.length, lock.cookie, lock.lockowner, lock.exclusive);
		if (!svc_sendreply(transp, (xdrproc_t)xdr_void, (Generic_t)NULL))
			if (debugmode)
				fprintf(stdout, "%s: [remove] Can't reply to RPC call\n", PROGNAME);
		return;
	case SHARE:
	case UNSHARE:
		if (!svc_getargs(transp, (xdrproc_t)xdr_share, (Generic_t)&lock)) {
			if (debugmode)
				fprintf(stdout, "%s: [%sshare] svc_getargs failed\n", PROGNAME, (rqstp->rq_proc==SHARE)? "": "un");
			svcerr_decode(transp);
			return;
		}
		lock.stat = !process_lock_request((int)rqstp->rq_proc, lock.name, lock.fh, lock.access, lock.mode, lock.offset, lock.length, lock.cookie, lock.lockowner, lock.exclusive);
		if (debugmode) {
			if (lock.stat)
				fprintf(stdout, "%s: [%sshare] lock status = false\n", PROGNAME, (rqstp->rq_proc==SHARE)? "": "un");
			else
				fprintf(stdout, "%s: [%sshare] lock status = true\n", PROGNAME, (rqstp->rq_proc==SHARE)? "": "un");
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_shareres, (Generic_t)&lock))
			if (debugmode)
				fprintf(stdout, "%s: [%sshare] Can't reply to RPC call\n", PROGNAME, (rqstp->rq_proc==SHARE)? "": "un");
		return;
	case ZAP_LOCKS:
		if (!svc_getargs(transp, (xdrproc_t)xdr_lock, (Generic_t)&lock)) {
			if (debugmode)
				fprintf(stdout, "%s: [zaplock] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		lock.stat = !process_lock_request((int) rqstp->rq_proc, lock.name, lock.fh, lock.access, lock.mode, lock.offset, lock.length, lock.cookie, lock.lockowner, lock.exclusive);
		if (debugmode) {
			if (lock.stat)
				fprintf(stdout, "%s: [zaplock] zaplock request failed\n", PROGNAME);
			else
				fprintf(stdout, "%s: [zaplock] zaplock request succeeded\n", PROGNAME);
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_res, (Generic_t)&lock))
			if (debugmode)
				fprintf(stdout, "%s: [zaplock] Can't reply to RPC call\n", PROGNAME);
		return;
	case LOCK:
		if (!svc_getargs(transp, (xdrproc_t)xdr_lock, (Generic_t)&lock)) {
			if (debugmode)
				fprintf(stdout, "%s: [lock] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		lock.stat = !process_lock_request((int) rqstp->rq_proc, lock.name, lock.fh, lock.access, lock.mode, lock.offset, lock.length, lock.cookie, lock.lockowner, lock.exclusive);
		if (debugmode) {
			if (lock.stat)
				fprintf(stdout, "%s: [lock] lock request failed\n", PROGNAME);
			else
				fprintf(stdout, "%s: [lock] lock request succeeded\n", PROGNAME);
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_res, (Generic_t)&lock))
			if (debugmode)
				fprintf(stdout, "%s: [lock] Can't reply to RPC call\n", PROGNAME);
		return;
	case UNLOCK:
		if (!svc_getargs(transp, (xdrproc_t)xdr_unlock, (Generic_t)&lock)) {
			if (debugmode)
				fprintf(stdout, "%s: [unlock] svc_getargs request failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		lock.stat = !process_lock_request((int) rqstp->rq_proc, lock.name, lock.fh, lock.access, lock.mode, lock.offset, lock.length, lock.cookie, lock.lockowner, lock.exclusive);
		if (debugmode) {
			if (lock.stat)
				fprintf(stdout, "%s: [unlock] unlock request failed\n", PROGNAME, (rqstp->rq_proc==LOCK)? "": (rqstp->rq_proc==UNLOCK)? "un": "zap");
			else
				fprintf(stdout, "%s: [unlock] unlock request succeeded\n", PROGNAME, (rqstp->rq_proc==LOCK)? "": (rqstp->rq_proc==UNLOCK)? "un": "zap");
		}
		if (!svc_sendreply(transp, (xdrproc_t)xdr_res, (Generic_t)&lock))
			if (debugmode)
				fprintf(stdout, "%s: [unlock] Can't reply to RPC call\n", PROGNAME);
		return;
	case GET_PRINTERS:
		if (!svc_sendreply(transp, (xdrproc_t)xdr_printers, (Generic_t)NULL))
			if (debugmode)
				fprintf(stdout, "%s: [get_printers] Can't reply to RPC call\n", PROGNAME);
		return;
	case GET_PRINTQ:
		
		if (!svc_getargs(transp, (xdrproc_t)xdr_get_printq, (Generic_t)&printq)) {
			if (debugmode)
				fprintf(stdout, "%s: [get_printq] svc_getargs failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		printq.printer_number = -1;
		for (i=0; i<num_printers; i++)
		{
			if (strcmp(printers[i].queuename, printq.queuename) == 0) {
				if (time(NULL) > (printers[i].lasttime+POLL_INTERVAL))
					add_entries(i);
				printq.printer_number = i;
				break;
			}
		}

		if (!svc_sendreply(transp, (xdrproc_t)xdr_send_printq, (Generic_t)&printq))
			if (debugmode)
				fprintf(stdout, "%s: [get_printq] Can't reply to RPC call\n", PROGNAME);
		return;
	case CANCEL_PRJOB:
		if (!svc_getargs(transp, (xdrproc_t)xdr_cancel_job, (Generic_t)&printq)) {
			if (debugmode)
				fprintf(stdout, "%s: [cancel_prjob] svc_getargs failed\n", PROGNAME);
			svcerr_decode(transp);
			return;
		}
		j = WN_BAD_QUEUE;
		for (i=0; i<num_printers; i++) {
			if (strcmp(printers[i].queuename, printq.queuename) == 0) {
				for (j=0; j<MAX_NUM_CANCEL_CACHE; j++) {
					if ((cancache[j].jobid == printq.printer_number) &&
					    (strcmp(cancache[j].username, printq.username) == 0) &&
					    (strcmp(cancache[j].queuename, printq.queuename) == 0)) {
						j = cancache[j].response;
						goto no_bloody_boolean_handy;
					}
				}
				if ((j=cancel_printjob(i, printq.username, printq.printer_number)) == WN_SUCCESS)
					printers[i].lasttime = 0L;
				cancache[current_cache].jobid = printq.printer_number;
				strcpy(cancache[current_cache].queuename, printq.queuename);
				strcpy(cancache[current_cache].username, printq.username);
				cancache[current_cache].response = j;
				current_cache = ++current_cache%MAX_NUM_CANCEL_CACHE;
				break;
			}
		}
no_bloody_boolean_handy:
		if (!svc_sendreply(transp, (xdrproc_t)xdr_int, (Generic_t)&j))
			if (debugmode)
				fprintf(stdout, "%s: [cancel_prjob] Can't reply to RPC call\n", PROGNAME);
		return;
	default:
		svcerr_noproc(transp);
		return;
	}
}

static struct fhstatus {
	int status;
	char dir[32];
} mount_return;

/*
 *	 Response from rpc.mountd for mount request of SPOOL directory.
 */
int
xdr_fhstatus(xdrsp, fhstat)
	XDR *xdrsp;
	struct fhstatus *fhstat;
{
unsigned int i,j;
char *p;

	if (debugmode)
		fprintf(stdout, "%s: [xdr_fhstatus] called\n", PROGNAME);
	if (!xdr_int(xdrsp, &fhstat->status))
		return(0);
	if (fhstat->status == 0) {
		for (i=0; i<32 ; i+=4) {
			if (!xdr_int(xdrsp, (int *)&j))
				return(0);
			*(_uint32 *)(&(fhstat->dir[0])+i) = j;
		}
	}
	if (debugmode > 1)
		fprintf(stdout, "%s: [xdr_fhstatus] returns Success\n", PROGNAME);
	return(1);
}

static struct text_string {
	int length;
	char *dir;
} request_text;

/*
 *	 Request to mount SPOOL directory.
 */
int
xdr_text(xdrsp, req)
	XDR *xdrsp;
	struct text_string *req;
{
	if (debugmode)
		fprintf(stdout, "%s: [xdr_text] mount request for %s\n", PROGNAME, req->dir);
	if (!xdr_bytes(xdrsp, &req->dir, (unsigned int *)&req->length, 255))
		return(0);
	if (debugmode > 1)
		fprintf(stdout, "%s: [xdr_text] returns Success\n", PROGNAME);
	return(1);
}


/*
 *	 Use rpc.mountd to determine the fhandle of the spool directory,
 *	 Can be replaced with extracted code from rpc.mountd.
 */
int get_mount(dir, directory)
	char *dir, *directory;
{
struct hostent *hp;
struct timeval pentry_timeout, total_timeout;
struct sockaddr_in server_addr;
int sock = RPC_ANYSOCK;
register CLIENT *client;
enum clnt_stat clnt_stat;

	if (debugmode)
		fprintf(stdout, "%s: [get_mount] called\n", PROGNAME);
	if (debugmode > 1)
		fprintf(stdout, "%s: [get_mount] dir = \"%s\", directory = \"%s\"\n", PROGNAME, dir, directory);
	request_text.length = strlen(directory);
	request_text.dir = directory;
	if ((hp=gethostbyname(hostname)) == NULL) {
		fprintf(stdout, "%s: [get_mount] unable to resolve machine name \"%s\" to an IP address.\n", PROGNAME, hostname);
		fprintf(stdout, "%s:             please contact your system administrator to fix this problem.\n", PROGNAME);
		fflush(stdout);
		exit(-1);
	}
	pentry_timeout.tv_sec = 3;
	pentry_timeout.tv_usec = 0;
	memcpy((caddr_t)&server_addr.sin_addr, hp->h_addr, hp->h_length);
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = 0;
	if ((client=(CLIENT *)clntudp_create(&server_addr, 100005L, 1L, pentry_timeout, &sock)) == NULL) {
		fprintf(stdout, "%s: [get_mount] clntupd_create failed\n", PROGNAME);
		clnt_pcreateerror("clntudp_create");
		fflush(stdout);
		exit(-1);
	}
	if (bindresvport(sock, (struct sockaddr_in *)NULL))
		if (debugmode)
			fprintf(stdout, "%s: [get_mount] cannot allocate reserved port ... not critical.\n", PROGNAME);
	client->cl_auth = authunix_create_default();
	total_timeout.tv_sec = 20;
	total_timeout.tv_usec = 0;
	clnt_stat = clnt_call(client, 1L, (xdrproc_t)xdr_text, (Generic_t)&request_text, (xdrproc_t)xdr_fhstatus, (Generic_t)&mount_return, total_timeout);
	if (clnt_stat != RPC_SUCCESS) {
		fprintf(stdout, "%s: [get_mount] clnt_stat failed\n", PROGNAME);
		clnt_perror(client,"rpc");
		fflush(stdout);
		exit(-1);
	}
	auth_destroy(client->cl_auth);
	clnt_destroy(client);
	if (mount_return.status == 0L)
		memcpy(dir, &mount_return.dir[0], 32);
	if (debugmode)
		fprintf(stdout, "%s: [get_mount] returns mount status %u\n", PROGNAME, mount_return.status);
	return(mount_return.status);
}


/*
 *	 Daemon should be killed with a simple KILL and not KILL -9 so
 *	 that it may unmap itself
 */
Void
die(unused)
	int unused;
{
	fprintf(stdout, "%s: [die] Daemon is dying\n\n", PROGNAME);
	fflush(stdout);

#ifndef TIRPC
	(void) pmap_unset(BW_NFS_PROGRAM, BW_NFS_VERSION);
#else
	svc_unreg(BW_NFS_PROGRAM, BW_NFS_VERSION);
#endif

	exit(1);
}

void
ErrAndDie(fmt, str1, str2, str3, str4)
	char *fmt, *str1, *str2, *str3, *str4;
{
	fprintf(stdout, fmt, str1, str2, str3, str4);
	fflush(stdout);
	exit(-1);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
SVCXPRT *transp;
char *p, *prog;
int i, child_pid;
struct hostent *hp;
#ifdef TIRPC
struct t_bind *req = NULL;
struct t_info tinfo;
struct netconfig *nconf = NULL;
#endif
int tfd = -1;
struct sockaddr_in serv_addr;
unsigned short bind_port = 0;

#ifdef SCO
	set_auth_parameters(argc, argv);
	setluid(0);
#endif

	prog = argv[0];
	debugmode = 0;
	debuglog = NULL;
	DontDoPrinting = 0;
	if (sizeof(_uint32) != 4)
		ErrAndDie("Integral size mismatch.  Try defining INT32_LONG and recompiling\n", NULL, NULL, NULL, NULL);
	fprintf(stdout, "%s Authentication Daemon Vsn %s starting\n\n", PROGNAME, VERSION);
	argc--;
	argv++;

another:
	if ((argc > 0) && (strcmp(*argv, "-A") == 0)) {
		argc--;
		argv++;
		A_flag = 1;
		goto another;
	}

	if ((argc > 0) && (strcmp(*argv, "-d") == 0)) {
		argc--;
		argv++;
		if (debuglog)
			ErrAndDie(USAGE, prog, NULL, NULL, NULL);
		debugmode++;
		goto another;
	}

	if ((argc > 0) && (strcmp(*argv, "-X") == 0)) {
		argc--;
		argv++;
		DontDoPrinting++;
		goto another;
	}

	if ((argc > 0) && (strcmp(*argv, "-s") == 0)) {
		argc--;
		argv++;
		if (argc < 2)
			ErrAndDie(USAGE, prog, NULL, NULL, NULL);
		read_ips(*argv);
		s_flag = 1;
		argc--;
		argv++;
		goto another;
	}
	if ((argc > 0) && (strcmp(*argv, "-l") == 0)) {
		argc--;
		argv++;
		if (argc < 1)
			ErrAndDie(USAGE, prog, NULL, NULL, NULL);
		if (debugmode)
			ErrAndDie(USAGE, prog, NULL, NULL, NULL);
		if (**argv)
			debuglog = my_strdup(*argv);
		if (debuglog == NULL)
			ErrAndDie("Out of Memory", NULL, NULL, NULL, NULL);
		if ((i=open(debuglog, O_RDWR|O_TRUNC|O_CREAT|O_NOCTTY)) < 0)
			ErrAndDie("exiting, Will not be able to write to logfile %s\n", debuglog, NULL, NULL, NULL);
		close(i);
		chmod(debuglog, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		fprintf(stdout, "Logging debug trace to %s\n", debuglog);
		argc--;
		argv++;
		goto another;
	}

	if ((argc > 0) && (strcmp(*argv, "-p") == 0)) {
		argc--;
		argv++;
		if (argc < 1)
			ErrAndDie(USAGE, prog, NULL, NULL, NULL);
		if (**argv)
			bind_port = atoi(*argv);
		if (bind_port == 0)
			ErrAndDie("Invalid port\n", NULL, NULL, NULL, NULL);
		argc--;
		argv++;
		goto another;
	}

	if ((argc != 1) && (!DontDoPrinting))
		ErrAndDie(USAGE, prog, NULL, NULL, NULL);

	init_locks();
	if ((hostname=(char *)malloc(255)) == NULL)
		ErrAndDie("%s: unable to allocate 255 bytes of storage; Terminating\n", PROGNAME, NULL, NULL, NULL);
	(void)gethostname(hostname, 255);
	if ((hp=gethostbyname(hostname)) == NULL)
		ErrAndDie("%s: This machine claims its name is \"%s\" but it cannot resolve that into an IP address\n%s: Please contact your system administrator to fix this problem and restart %s\n", PROGNAME, hostname, PROGNAME, prog);
	memcpy((caddr_t) &my_ip, hp->h_addr, hp->h_length);
	if (!DontDoPrinting) {
		if (get_mount(dir, *argv) != 0)
			ErrAndDie("%s: Spool area mount point %s could not be mounted.\n%s: Check that %s is in the /etc/exports file & world-writable\n", PROGNAME, *argv, PROGNAME, *argv);
	}
	else
		if (debugmode)
			fprintf(stdout, "%s: [main] Daemon will not accept print requests.\n", PROGNAME);

	if (!debugmode) {
#ifdef SIGTTOU
		signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
		signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
		signal(SIGTSTP, SIG_IGN);
#endif
		if ((child_pid=fork()) > 0)
			exit(0);
		if (child_pid < 0)
			ErrAndDie("%s: Unable to detach process\n", PROGNAME, NULL, NULL, NULL);
#ifdef SYSV_PGRP
		if (setpgrp() == -1)
#else
		if (setpgrp(0, 0) == -1)
#endif
			ErrAndDie("%s: [main] Can't change process group\n", PROGNAME, NULL, NULL, NULL);
#ifdef SIGTSTP
#ifdef TIOCNOTTY
		if ((i=open("/dev/tty", O_RDWR)) >= 0) {
			ioctl(i, TIOCNOTTY, NULL);
			close(i);
		}
#endif
#else
		signal(SIGHUP, SIG_IGN);
		if ((child_pid=fork()) > 0)
			exit(0);
		if (child_pid < 0)
			ErrAndDie("%s: [main] Unable to detach second process\n", PROGNAME, NULL, NULL, NULL);
#endif
#ifndef NOFILE
#define NOFILE 10
#endif
#ifdef LIBS_USE_OPEN_FILES
#undef NOFILE
#define NOFILE	3
#endif
		for (i=0; i<NOFILE; i++)
			close(i);
		errno = 0;
		chdir("/");
		umask(0);

		if ((i=open("/dev/null", O_RDWR)) < 0)
			ErrAndDie("%s: [main] Unable to map handles\n", PROGNAME, NULL, NULL, NULL);
		dup2(i, 4);
		close(i);
		dup2(4, 0);
		dup2(4, 1);
		dup2(4, 2);
		close(4);
		if (debuglog && *debuglog) {
			if ((i=open(debuglog,O_RDWR|O_APPEND|O_CREAT|O_NOCTTY)) < 0)
				ErrAndDie("Could not open logfile %s", debuglog, NULL, NULL, NULL);

			dup2(i, 1);
			dup2(i, 2);
			fprintf(stdout, "%s: [main] real user %d, effective user %d\n", PROGNAME, getuid(), geteuid());
			fprintf(stdout,"%s: [main] real group %d, effective group %d\n", PROGNAME, getgid(), getegid());
			debugmode++;
		}
	}

	if (!DontDoPrinting) {
		if (determine_print_command() == 0)
			ErrAndDie("%s: [main] Unable to determine print command\n", PROGNAME, NULL, NULL, NULL);
		p = strcpy(spool_dir, *argv) + strlen(*argv);
		if (*(p-1) == '/')
			*(--p) = '\0';
		f.status = 0;
		f.dir_size = 32;
		f.dir_handle = dir;
	}
	else
		f.status = 1;

	get_printers();
	get_groups();

	/* Bind HCLNFSD server to specified port */
	if (bind_port > 0)
	{
		memset(&serv_addr, 0, sizeof(serv_addr));

		serv_addr.sin_family = AF_INET;
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		serv_addr.sin_port = htons(bind_port);
	}

#ifndef TIRPC
	/* BSD sockets */
	tfd = RPC_ANYSOCK;

	if (bind_port > 0)
	{
		if ((tfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
			ErrAndDie("%s: [main] Unable to create socket (errno=%d)\n",
					PROGNAME, errno, NULL, NULL);

		/* Attempt to bind socket to requested port */
		if (bind(tfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
			ErrAndDie("%s: [main] Unable to bind HCLNFSD to port %d\n",
					PROGNAME, bind_port, NULL, NULL);
	}

	transp = (SVCXPRT *) svcudp_create(tfd);
#else
	/* TI-RPC */
	tfd = RPC_ANYFD;
	
	if ((nconf = getnetconfigent("udp")) == NULL)
		ErrAndDie("%s: [main] Unable to get netconfig record (t_errno=%d)\n",
				PROGNAME, t_errno, NULL, NULL);

	if ((tfd = t_open(nconf->nc_device, O_RDWR, &tinfo)) < 0)
		ErrAndDie("%s: [main] Unable to open %s (t_errno=%d)\n",
				PROGNAME, nconf->nc_device, t_errno, NULL, NULL);

	if (bind_port > 0)
	{
		if ((req = (struct t_bind *) t_alloc(tfd, T_BIND, T_ADDR)) == NULL)
			ErrAndDie("%s: [main] t_alloc() failed (t_errno=%d)\n",
					PROGNAME, t_errno, NULL, NULL);

		req->addr.maxlen = sizeof(serv_addr);
		req->addr.len = sizeof(serv_addr);
		req->addr.buf = (char *) &serv_addr;

		req->qlen = 5;

		if (t_bind(tfd, req, NULL) < 0)
		{
			t_free(req, T_BIND);

			ErrAndDie("%s: [main] Unable to bind to port %d (t_errno=%d)\n",
					PROGNAME, bind_port, t_errno, NULL);
		}
	}

	transp = svc_tli_create(tfd, nconf, req, 0, 0);

	if ((transp == NULL) && (bind_port > 0))
	{
		/* svc_tli_create() failed */
		t_free(req, T_BIND);
		t_close(tfd);
	}
#endif

	if (transp == NULL)
		ErrAndDie("%s: [main] Unable to create an RPC server (errno=%d)\n", 
				PROGNAME, errno, NULL, NULL);

	/* Register HCLNFSD with portmapper */
#ifndef TIRPC
	(void) pmap_unset(BW_NFS_PROGRAM, BW_NFS_VERSION);

	if (!svc_register(transp, BW_NFS_PROGRAM, BW_NFS_VERSION,
				(void *) create_spool, IPPROTO_UDP))
		ErrAndDie("%s: [main] Can't register HCLNFSD service\n", 
				PROGNAME, NULL, NULL, NULL);
#else
	if (bind_port > 0)
		transp->xp_port = serv_addr.sin_port;

	svc_unreg(BW_NFS_PROGRAM, BW_NFS_VERSION);

	if (!svc_reg(transp, BW_NFS_PROGRAM, BW_NFS_VERSION, create_spool, nconf))
	{
		ErrAndDie("%s: [main] Can't register HCLNFSD service (t_errno=%d)\n",
				PROGNAME, t_errno, NULL, NULL);
	}
#endif

	if ((debugmode) && (bind_port > 0))
		fprintf(stdout, "%s: [main] bound to port %d\n", 
				PROGNAME, bind_port);

	(void) signal(SIGHUP, clear);
	(void) signal(SIGINT, die);
	(void) signal(SIGQUIT, die);
#ifdef SIGABRT
	(void) signal(SIGABRT, die);
#endif
	(void) signal(SIGTERM, die);
#ifdef SIGCHLD
	(void) signal(SIGCHLD, free_child);
#endif
#ifdef SIGCLD
/*
 * I don't _think_ this change will hurt anything (it should just make
 * non-bsd systems wait for children rather than ignoring return codes).
 *
 *#ifndef SIGTSTP
 *	(void) signal(SIGCLD, SIG_IGN);
 *#else
 */
	(void) signal(SIGCLD, free_child);
/*#endif*/
#endif
	svc_run();
	fprintf(stdout, "%s: [main] svc_run returned!\n", PROGNAME);
	fflush(stdout);
	exit(1);
	return(1); /* for the compulsive compilers */
}
